"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__

hub.exec.boto3.client.apigatewayv2.create_integration
hub.exec.boto3.client.apigatewayv2.delete_integration
hub.exec.boto3.client.apigatewayv2.get_apis
hub.exec.boto3.client.apigatewayv2.get_integration
hub.exec.boto3.client.apigatewayv2.get_integrations
hub.exec.boto3.client.apigatewayv2.update_integration
"""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    api_id: str,
    integration_type: str,
    resource_id: str = None,
    connection_id: str = None,
    connection_type: str = None,
    content_handling_strategy: str = None,
    credentials_arn: str = None,
    description: str = None,
    integration_method: str = None,
    integration_subtype: str = None,
    integration_uri: str = None,
    passthrough_behavior: str = None,
    payload_format_version: str = None,
    request_parameters: Dict[str, str] = None,
    request_templates: Dict[str, str] = None,
    response_parameters: Dict[str, Dict[str, str]] = None,
    template_selection_expression: str = None,
    timeout_in_millis: int = None,
    tls_config: make_dataclass(
        "TlsConfigInput", [("ServerNameToVerify", str, field(default=None))]
    ) = None,
) -> Dict[str, Any]:
    r"""
    Creates an API Gateway v2 integration resource.

    Args:
        name(string): An Idem name of the resource.
        api_id(string): The API resource identifier in Amazon Web Services.
        integration_type(string): The integration type of an integration. One of the following: AWS: for integrating the route or
            method request with an AWS service action, including the Lambda function-invoking action. With
            the Lambda function-invoking action, this is referred to as the Lambda custom integration. With
            any other AWS service action, this is known as AWS integration. Supported only for WebSocket
            APIs. AWS_PROXY: for integrating the route or method request with a Lambda function or other AWS
            service action. This integration is also referred to as a Lambda proxy integration. HTTP: for
            integrating the route or method request with an HTTP endpoint. This integration is also referred
            to as the HTTP custom integration. Supported only for WebSocket APIs. HTTP_PROXY: for
            integrating the route or method request with an HTTP endpoint, with the client request passed
            through as-is. This is also referred to as HTTP proxy integration. For HTTP API private
            integrations, use an HTTP_PROXY integration. MOCK: for integrating the route or method request
            with API Gateway as a "loopback" endpoint without invoking any backend. Supported only for
            WebSocket APIs.
        resource_id(string, optional): The integration resource identifier in Amazon Web Services.
        connection_id(string, optional): The ID of the VPC link for a private integration. Supported only for HTTP APIs. Defaults to None.
        connection_type(string, optional): The type of the network connection to the integration endpoint. Specify INTERNET for connections
            through the public routable internet or VPC_LINK for private connections between API Gateway and
            resources in a VPC. The default value is INTERNET.
        content_handling_strategy(string, optional): Supported only for WebSocket APIs. Specifies how to handle response payload content type
            conversions. Supported values are CONVERT_TO_BINARY and CONVERT_TO_TEXT, with the following
            behaviors: CONVERT_TO_BINARY: Converts a response payload from a Base64-encoded string to the
            corresponding binary blob. CONVERT_TO_TEXT: Converts a response payload from a binary blob to a
            Base64-encoded string. If this property is not defined, the response payload will be passed
            through from the integration response to the route response or method response without
            modification.
        credentials_arn(string, optional): Specifies the credentials required for the integration, if any. For AWS integrations, three
            options are available. To specify an IAM Role for API Gateway to assume, use the role's Amazon
            Resource Name (ARN). To require that the caller's identity be passed through from the request,
            specify the string arn:aws:iam::*:user/*. To use resource-based permissions on supported AWS
            services, specify null.
        description(string, optional): The description of the integration.
        integration_method(string, optional): Specifies the integration's HTTP method type. Defaults to None.
        integration_subtype(string, optional): Supported only for HTTP API AWS_PROXY integrations. Specifies the AWS service action to invoke.
            To learn more, see Integration subtype reference.
        integration_uri(string, optional): For a Lambda integration, specify the URI of a Lambda function. For an HTTP integration, specify
            a fully-qualified URL. For an HTTP API private integration, specify the ARN of an Application
            Load Balancer listener, Network Load Balancer listener, or AWS Cloud Map service. If you specify
            the ARN of an AWS Cloud Map service, API Gateway uses DiscoverInstances to identify resources.
            You can use query parameters to target specific resources. To learn more, see DiscoverInstances.
            For private integrations, all resources must be owned by the same AWS account.
        passthrough_behavior(string, optional): Specifies the pass-through behavior for incoming requests based on the Content-Type header in
            the request, and the available mapping templates specified as the requestTemplates property on
            the Integration resource. There are three valid values: WHEN_NO_MATCH, WHEN_NO_TEMPLATES, and
            NEVER. Supported only for WebSocket APIs. WHEN_NO_MATCH passes the request body for unmapped
            content types through to the integration backend without transformation. NEVER rejects unmapped
            content types with an HTTP 415 Unsupported Media Type response. WHEN_NO_TEMPLATES allows pass-
            through when the integration has no content types mapped to templates. However, if there is at
            least one content type defined, unmapped content types will be rejected with the same HTTP 415
            Unsupported Media Type response.
        payload_format_version(string, optional): Specifies the format of the payload sent to an integration. Required for HTTP APIs. Defaults to None.
        request_parameters(Dict, optional): For WebSocket APIs, a key-value map specifying request parameters that are passed from the
            method request to the backend. The key is an integration request parameter name and the
            associated value is a method request parameter value or static value that must be enclosed
            within single quotes and pre-encoded as required by the backend. The method request parameter
            value must match the pattern of method.request.{location}.{name}, where
            {location} is querystring, path, or header; and {name}
            must be a valid and unique method request parameter name. For HTTP API integrations with a
            specified integrationSubtype, request parameters are a key-value map specifying parameters that
            are passed to AWS_PROXY integrations. You can provide static values, or map request data, stage
            variables, or context variables that are evaluated at runtime. To learn more, see Working with
            AWS service integrations for HTTP APIs. For HTTP API integrations without a specified
            integrationSubtype request parameters are a key-value map specifying how to transform HTTP
            requests before sending them to the backend. The key should follow the pattern
            <action>:<header|querystring|path>.<location> where action can be append, overwrite or remove.
            For values, you can provide static values, or map request data, stage variables, or context
            variables that are evaluated at runtime. To learn more, see Transforming API requests and
            responses.
        request_templates(Dict, optional): Represents a map of Velocity templates that are applied on the request payload based on the
            value of the Content-Type header sent by the client. The content type value is the key in this
            map, and the template (as a String) is the value. Supported only for WebSocket APIs.
        response_parameters(Dict, optional): Supported only for HTTP APIs. You use response parameters to transform the HTTP response from a
            backend integration before returning the response to clients. Specify a key-value map from a
            selection key to response parameters. The selection key must be a valid HTTP status code within
            the range of 200-599. Response parameters are a key-value map. The key must match pattern
            <action>:<header>.<location> or overwrite.statuscode. The action can be append, overwrite or
            remove. The value can be a static value, or map to response data, stage variables, or context
            variables that are evaluated at runtime. To learn more, see Transforming API requests and
            responses.
        template_selection_expression(string, optional): The template selection expression for the integration.
        timeout_in_millis(int, optional): Custom timeout between 50 and 29,000 milliseconds for WebSocket APIs and between 50 and 30,000
            milliseconds for HTTP APIs. The default timeout is 29 seconds for WebSocket APIs and 30 seconds
            for HTTP APIs.
        tls_config(Dict[str, Any], optional): The TLS configuration for a private integration. If you specify a TLS configuration, private
            integration traffic uses the HTTPS protocol. Supported only for HTTP APIs. Defaults to None.
            * ServerNameToVerify (str, optional): If you specify a server name, API Gateway uses it to verify the hostname on the integration's
                certificate. The server name is also included in the TLS handshake to support Server Name
                Indication (SNI) or virtual hosting.

    Request Syntax:
        [idem_test_aws_apigatewayv2_route]:
          aws.apigatewayv2.route.present:
            - name: 'string'
            - api_id: 'string'
            - integration_type: 'AWS'|'HTTP'|'MOCK'|'HTTP_PROXY'|'AWS_PROXY'
            - connection_id: 'string'
            - connection_type: 'INTERNET'|'VPC_LINK'
            - content_handling_strategy: 'AWS'|'HTTP'|'MOCK'|'HTTP_PROXY'|'AWS_PROXY'
            - credentials_arn: 'string'
            - description: 'string'
            - integration_method: 'string'
            - integration_subtype: 'string'
            - integration_uri: 'string'
            - passthrough_behavior: 'WHEN_NO_MATCH'|'NEVER'|'WHEN_NO_TEMPLATES'
            - payload_format_version: 'string'
            - request_parameters: {'string': 'string'}
            - request_templates: {'string': 'string'}
            - response_parameters: {string': {'string': 'string'}}
            - template_selection_expression: 'string'
            - timeout_in_millis: int
            - tls_config: {'ServerNameToVerify': 'string'}

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            idem_test_aws_apigatewayv2_route:
              aws.apigatewayv2.route.present:
                - name: value
                - api_id: value
                - integration_type: value
                - connection_id: value
                - connection_type: value
                - content_handling_strategy: value
                - credentials_arn: value
                - description: value
                - integration_method: value
                - integration_subtype: value
                - integration_uri: value
                - passthrough_behavior: value
                - payload_format_version: value
                - request_parameters: {}
                - request_templates: {}
                - response_parameters: {}
                - template_selection_expression: value
                - timeout_in_millis: value
                - tls_config: {}
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    resource_updated = False

    if resource_id:
        before_ret = await hub.exec.aws.apigatewayv2.integration.get(
            ctx=ctx, name=name, resource_id=resource_id, api_id=api_id
        )
        if not before_ret["result"] or not before_ret["ret"]:
            result["result"] = False
            result["comment"] = tuple(before_ret["comment"])
            return result

        result["comment"] = hub.tool.aws.comment_utils.already_exists_comment(
            resource_type="aws.apigatewayv2.integration", name=name
        )
        result["old_state"] = copy.deepcopy(before_ret["ret"])
        result["new_state"] = copy.deepcopy(result["old_state"])

        resource_parameters = {
            "connection_id": connection_id,
            "connection_type": connection_type,
            "content_handling_strategy": content_handling_strategy,
            "credentials_arn": credentials_arn,
            "description": description,
            "integration_method": integration_method,
            "integration_subtype": integration_subtype,
            "integration_type": integration_type,
            "integration_uri": integration_uri,
            "passthrough_behavior": passthrough_behavior,
            "payload_format_version": payload_format_version,
            "request_parameters": request_parameters,
            "request_templates": request_templates,
            "response_parameters": response_parameters,
            "template_selection_expression": template_selection_expression,
            "timeout_in_millis": timeout_in_millis,
            "tls_config": tls_config,
        }

        update_integration_ret = await hub.tool.aws.apigatewayv2.integration.update(
            ctx,
            api_id=api_id,
            resource_id=resource_id,
            raw_resource=before_ret["ret"],
            resource_parameters=resource_parameters,
        )
        result["comment"] = result["comment"] + update_integration_ret["comment"]
        if not update_integration_ret["result"]:
            result["result"] = False
            return result

        resource_updated = resource_updated or bool(update_integration_ret["ret"])
        if update_integration_ret["ret"] and ctx.get("test", False):
            result["new_state"].update(update_integration_ret["ret"])

        if resource_updated and ctx.get("test", False):
            return result
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "api_id": api_id,
                    "connection_id": connection_id,
                    "connection_type": connection_type,
                    "content_handling_strategy": content_handling_strategy,
                    "credentials_arn": credentials_arn,
                    "description": description,
                    "integration_method": integration_method,
                    "integration_subtype": integration_subtype,
                    "integration_type": integration_type,
                    "integration_uri": integration_uri,
                    "passthrough_behavior": passthrough_behavior,
                    "payload_format_version": payload_format_version,
                    "request_parameters": request_parameters,
                    "request_templates": request_templates,
                    "response_parameters": response_parameters,
                    "template_selection_expression": template_selection_expression,
                    "timeout_in_millis": timeout_in_millis,
                    "tls_config": tls_config,
                },
            )
            result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                resource_type="aws.apigatewayv2.integration", name=name
            )
            return result

        create_ret = await hub.exec.boto3.client.apigatewayv2.create_integration(
            ctx,
            ApiId=api_id,
            ConnectionId=connection_id,
            ConnectionType=connection_type,
            ContentHandlingStrategy=content_handling_strategy,
            CredentialsArn=credentials_arn,
            Description=description,
            IntegrationMethod=integration_method,
            IntegrationSubtype=integration_subtype,
            IntegrationType=integration_type,
            IntegrationUri=integration_uri,
            PassthroughBehavior=passthrough_behavior,
            PayloadFormatVersion=payload_format_version,
            RequestParameters=request_parameters,
            RequestTemplates=request_templates,
            ResponseParameters=response_parameters,
            TemplateSelectionExpression=template_selection_expression,
            TimeoutInMillis=timeout_in_millis,
            TlsConfig=tls_config,
        )
        if not create_ret["result"]:
            result["result"] = False
            result["comment"] = create_ret["comment"]
            return result

        result["comment"] = hub.tool.aws.comment_utils.create_comment(
            resource_type="aws.apigatewayv2.integration", name=name
        )
        resource_id = create_ret["ret"]["IntegrationId"]

    if (not result["old_state"]) or resource_updated:
        after_ret = await hub.exec.aws.apigatewayv2.integration.get(
            ctx=ctx, name=name, resource_id=resource_id, api_id=api_id
        )
        if not after_ret["result"] or not after_ret["ret"]:
            result["result"] = False
            result["comment"] = result["comment"] + tuple(after_ret["comment"])
            return result
        result["new_state"] = after_ret["ret"]

    return result


async def absent(
    hub, ctx, name: str, api_id: str, resource_id: str = None
) -> Dict[str, Any]:
    r"""
    Deletes an API Gateway v2 integration resource.

    Args:
        name(string): An Idem name of the resource.
        api_id(string): The API resource identifier in Amazon Web Services.
        resource_id(string, optional): The Integration resource identifier in Amazon Web Services.
            Idem automatically considers this resource being absent if this field is not specified.

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            idem_test_aws_apigatewayv2_integration:
              aws.apigatewayv2.integration.absent:
                - name: value
                - api_id: value
                - resource_id: value
    """

    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.apigatewayv2.integration", name=name
        )
        return result

    before_ret = await hub.exec.aws.apigatewayv2.integration.get(
        ctx=ctx, name=name, resource_id=resource_id, api_id=api_id
    )
    if not before_ret["result"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.apigatewayv2.integration", name=name
        )
        return result
    if not before_ret["ret"]:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.apigatewayv2.integration", name=name
        )
    elif ctx.get("test", False):
        result["old_state"] = before_ret["ret"]
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type="aws.apigatewayv2.integration", name=name
        )
        return result
    else:
        result["old_state"] = before_ret["ret"]
        delete_ret = await hub.exec.boto3.client.apigatewayv2.delete_integration(
            ctx, ApiId=api_id, IntegrationId=resource_id
        )
        if not delete_ret["result"]:
            result["result"] = False
            result["comment"] = delete_ret["comment"]
            return result

        result["comment"] = hub.tool.aws.comment_utils.delete_comment(
            resource_type="aws.apigatewayv2.integration", name=name
        )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    Describe the resource in a way that can be recreated/managed with the corresponding "present" function

    Gets the API Gateway v2 integration resources for an AWS account.

    Returns:
        Dict[str, Dict[str, Any]]

    Examples:

        .. code-block:: bash

            $ idem describe aws.apigatewayv2.integration
    """

    result = {}

    get_apis_ret = await hub.exec.boto3.client.apigatewayv2.get_apis(ctx)
    if not get_apis_ret["result"]:
        hub.log.debug(f"Could not get apis {get_apis_ret['comment']}")
        return result

    for api in get_apis_ret["ret"]["Items"]:
        api_id = api.get("ApiId")

        get_integrations_ret = (
            await hub.exec.boto3.client.apigatewayv2.get_integrations(ctx, ApiId=api_id)
        )
        if not get_integrations_ret["result"]:
            hub.log.debug(
                f"Could not get integrations for ApiId '{api_id}': "
                f"{get_integrations_ret['comment']}. Describe will skip this api and continue."
            )
            continue

        for integration in get_integrations_ret["ret"]["Items"]:
            resource_translated = hub.tool.aws.apigatewayv2.integration.convert_raw_integration_to_present(
                api_id=api_id,
                raw_resource=integration,
            )

            result[resource_translated["name"]] = {
                "aws.apigatewayv2.integration.present": [
                    {parameter_key: parameter_value}
                    for parameter_key, parameter_value in resource_translated.items()
                ]
            }

    return result


async def search(hub, ctx, name, api_id: str, resource_id: str) -> Dict[str, Any]:
    """
    Use an un-managed API Gateway v2 integration resource as a data-source. This function has been deprecated.
    Please use exec.run with aws.apigatewayv2.integration.get instead.

    Args:
        name(string): An Idem name of the resource.
        api_id(string): The API resource identifier in Amazon Web Services.
        resource_id(string): The Integration resource identifier in Amazon Web Services.

    Request syntax:
        [idem_test_aws_apigatewayv2_integration]:
          aws.apigatewayv2.integration.search:
            - api_id: 'string'
            - resource_id: 'string'

    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: sls

            idem_test_aws_apigatewayv2_integration:
              aws.apigatewayv2.integration.search:
                - api_id: value
                - resource_id: value
    """
    hub.log.warning(
        f"aws.apigatewayv2.integration.search '{name}' state has been deprecated. Please use exec.run with aws.apigatewayv2.integration.get instead."
    )
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    ret = await hub.exec.aws.apigatewayv2.integration.get(
        ctx=ctx, name=name, resource_id=resource_id, api_id=api_id
    )
    if not ret["result"]:
        result["old_state"] = ret["ret"]
        result["new_state"] = copy.deepcopy(result["old_state"])
    return result



"""State module for managing Amazon Cloudtrail Trail resource."""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    s3_bucket_name: str,
    resource_id: str = None,
    s3_key_prefix: str = None,
    sns_topic_name: str = None,
    include_global_service_events: bool = None,
    is_multi_region_trail: bool = None,
    enable_logfile_validation: bool = None,
    cloud_watch_logs_loggroup_arn: str = None,
    cloud_watch_logs_role_arn: str = None,
    kms_key_id: str = None,
    is_organization_trail: bool = None,
    tags: Dict[str, Any]
    or List[
        make_dataclass(
            "Tag",
            [("Key", str, field(default=None)), ("Value", str, field(default=None))],
        )
    ] = None,
    is_logging: bool = None,
    insight_selectors: List[
        make_dataclass("InsightSelectors", [("InsightType", str, field(default=None))])
    ] = None,
    event_selectors: List[
        make_dataclass(
            "EventSelector",
            [
                ("ReadWriteType", str, field(default=None)),
                ("IncludeManagementEvents", bool, field(default=None)),
                (
                    "DataResources",
                    List[
                        make_dataclass(
                            "DataResource",
                            [
                                ("Type", str, field(default=None)),
                                ("Values", List[str], field(default=None)),
                            ],
                        )
                    ],
                    field(default=None),
                ),
                ("ExcludeManagementEventSources", List[str], field(default=None)),
            ],
        )
    ] = None,
    advanced_event_selectors: List[
        make_dataclass(
            "AdvancedEventSelector",
            [
                (
                    "FieldSelectors",
                    List[
                        make_dataclass(
                            "AdvancedFieldSelector",
                            [
                                ("Field", str),
                                ("Equals", List[str], field(default=None)),
                                ("StartsWith", List[str], field(default=None)),
                                ("EndsWith", List[str], field(default=None)),
                                ("NotEquals", List[str], field(default=None)),
                                ("NotStartsWith", List[str], field(default=None)),
                                ("NotEndsWith", List[str], field(default=None)),
                            ],
                        )
                    ],
                ),
                ("Name", str, field(default=None)),
            ],
        )
    ] = None,
) -> Dict[str, Any]:
    """Creates a trail that specifies the settings for delivery of log data to an Amazon S3 bucket.

    Args:
        name (str):
            An Idem name of the resource
            Specifies the name of the trail. The name must meet the following requirements:

            * Contain only ASCII letters (a-z, A-Z), numbers (0-9), periods (.), underscores (_), or dashes (-)
            * Start with a letter or number, and end with a letter or number
            * Be between 3 and 128 characters
            * Have no adjacent periods, underscores or dashes. Names like my-_namespace and my--namespace are not valid.
            * Not be in IP address format (for example, 192.168.5.4)

        s3_bucket_name (str):
            Specifies the name of the Amazon S3 bucket designated for publishing log files. See Amazon S3 Bucket
            Naming Requirements.

        resource_id (str, Optional):
            Trail name to identify the resource

        s3_key_prefix (str, Optional):
            Specifies the Amazon S3 key prefix that comes after the name of the bucket you have
            designated for log file delivery. For more information, see Finding Your CloudTrail Log Files .
            The maximum length is 200 characters.

        sns_topic_name (str, Optional) :
            Specifies the name of the Amazon SNS topic defined for notification of log file
            delivery. The maximum length is 256 characters.

        include_global_service_events (bool, Optional):
            Specifies whether the trail is publishing events from global services  such as IAM to the log files.

        is_multi_region_trail (bool, Optional) :
            Specifies whether the trail is created in the current region or in all regions.
            The default is false, which creates a trail only in the region where you are signed in. As a best practice,
            consider creating trails that log events in all regions.

        enable_logfile_validation (bool, Optional):
            Specifies whether log file integrity validation is enabled. The default is false.

        cloud_watch_logs_loggroup_arn (str, Optional) :
            Specifies a log group name using an Amazon Resource Name (ARN), a unique
            identifier that represents the log group to which CloudTrail logs will be delivered. Not required unless you
            specify CloudWatchLogsRoleArn .

        cloud_watch_logs_role_arn (str, Optional) :
            Specifies the role for the CloudWatch Logs endpoint to assume to write to a user's log group.

        kms_key_id (str, Optional):
            Specifies the KMS key ID to use to encrypt the logs delivered by CloudTrail. The value can
            be an alias name prefixed by "alias/", a fully specified ARN to an alias, a fully specified ARN to a key, or
            a globally unique identifier.

            CloudTrail also supports KMS multi-Region keys. For more information about multi-Region keys, see Using
            multi-Region keys in the Key Management Service Developer Guide.

            Examples of multi-Region keys:

            * alias/MyAliasName
            * arn:aws:kms:us-east-2:123456789012:alias/MyAliasName
            * arn:aws:kms:us-east-2:123456789012:key/12345678-1234-1234-1234-123456789012
            * 12345678-1234-1234-1234-123456789012

        is_organization_trail (bool, Optional):
             Specifies whether the trail is created for all accounts in an organization in
             Organizations, or only for the current Amazon Web Services account. The default is false,
             and cannot be true unless the call is made on behalf of an Amazon Web Services account
             that is the management account for an organization in Organizations.

        tags (dict or list, Optional):
            Dict in the format of {tag-key: tag-value} or List of tags in the format of
            [{"Key": tag-key, "Value": tag-value}] to associate with the CloudTrail trail.

            * (Key): The key in a key-value pair. The key must be must be no longer than 128 Unicode characters.
              The key must be unique for the resource to which it applies.
            * (Value, Optional): The value in a key-value pair of a tag.
              The value must be no longer than 256 Unicode characters.

        is_logging (bool, Optional) :
            Start and Stop the logging of CloudTrail

        insight_selectors(list[dict[str, Any]]):
            A JSON string that contains the insight types you want to log on a trail.
            ApiCallRateInsight and ApiErrorRateInsight are valid insight types.

            * InsightType (str, Optional):
              The type of insights to log on a trail. ApiCallRateInsight and ApiErrorRateInsight are valid
              insight types.

        event_selectors(list[dict[str, Any]], Optional):
            Specifies the settings for your event selectors. You can configure up to five event selectors
            for a trail. You can use either EventSelectors or AdvancedEventSelectors in a PutEventSelectors
            request, but not both. If you apply EventSelectors to a trail, any existing
            AdvancedEventSelectors are overwritten. Defaults to None.

            * ReadWriteType (str, Optional):
              Specify if you want your trail to log read-only events, write-only events, or all. For example,
              the EC2 GetConsoleOutput is a read-only API operation and RunInstances is a write-only API
              operation. By default, the value is All.

            * IncludeManagementEvents (bool, Optional):
              Specify if you want your event selector to include management events for your trail.
              For more information, see Management Events in the CloudTrail User Guide. By default, the value is true.
              The first copy of management events is free. You are charged for additional copies of management
              events that you are logging on any subsequent trail in the same region. For more information
              about CloudTrail pricing, see CloudTrail Pricing.

            * DataResources (list[dict[str, Any]], Optional):
              CloudTrail supports data event logging for Amazon S3 objects, Lambda functions, and Amazon
              DynamoDB tables with basic event selectors. You can specify up to 250 resources for an
              individual event selector, but the total number of data resources cannot exceed 250 across all
              event selectors in a trail. This limit does not apply if you configure resource logging for all
              data events.
              For more information, see Data Events and Limits in CloudTrail in the CloudTrail User Guide.

              * Type (str, Optional): The resource type in which you want to log data events.
                You can specify the following basic event selector resource types:

                * AWS::S3::Object
                * AWS::Lambda::Function
                * AWS::DynamoDB::Table

                The following resource types are also availble through advanced event selectors.
                Basic event selector resource types are valid in advanced event selectors,
                but advanced event selector resource types are not valid in basic event selectors.
                For more information, see AdvancedFieldSelector$Field.

              * Values (list[str], Optional):
                An array of Amazon Resource Name (ARN) strings or partial ARN strings for the specified objects.

                * To log data events for all objects in all S3 buckets in your Amazon Web Services account,
                  specify the prefix as arn:aws:s3:::.
                  This also enables logging of data event activity performed by any user or role in your
                  Amazon Web Services account, even if that activity is performed on a
                  bucket that belongs to another Amazon Web Services account.

                * To log data events for all objects in an S3 bucket, specify the bucket and an empty object
                  prefix such as arn:aws:s3:::bucket-1/. The trail logs data events for all objects in this S3 bucket.

                * To log data events for specific objects, specify the S3 bucket and object prefix such as
                  arn:aws:s3:::bucket-1/example-images. The trail logs data events for objects in this S3 bucket
                  that match the prefix.

                * To log data events for all Lambda functions in your Amazon Web Services
                  account, specify the prefix as arn:aws:lambda.  This also enables logging of Invoke activity
                  performed by any user or role in your Amazon Web Services account, even if that activity is
                  performed on a function that belongs to another Amazon Web Services account.

                * To log data events for a specific Lambda function, specify the function ARN.
                  Lambda function ARNs are exact. For example, if you specify a function ARN arn:aws:lambda:us-
                  west-2:111111111111:function:helloworld, data events will only be logged for arn:aws:lambda:us-
                  west-2:111111111111:function:helloworld. They will not be logged for arn:aws:lambda:us-
                  west-2:111111111111:function:helloworld2.

                * To log data events for all DynamoDB tables in your Amazon Web Services account,
                  specify the prefix as arn:aws:dynamodb.

            * ExcludeManagementEventSources (list[str], Optional):
              An optional list of service event sources from which you do not want management events to be
              logged on your trail. In this release, the list can be empty (disables the filter), or it can
              filter out Key Management Service or Amazon RDS Data API events by containing kms.amazonaws.com
              or rdsdata.amazonaws.com. By default, ExcludeManagementEventSources is empty, and KMS and Amazon
              RDS Data API events are logged to your trail. You can exclude management event sources only in
              regions that support the event source.

        advanced_event_selectors(list[dict[str, Any]], Optional):
            Specifies the settings for advanced event selectors. You can add advanced event selectors, and
            conditions for your advanced event selectors, up to a maximum of 500 values for all conditions
            and selectors on a trail. You can use either AdvancedEventSelectors or EventSelectors, but not
            both. If you apply AdvancedEventSelectors to a trail, any existing EventSelectors are
            overwritten. For more information about advanced event selectors, see Logging data events for
            trails in the CloudTrail User Guide. Defaults to None.

            * Name (str, Optional): An optional, descriptive name for an advanced event selector,
              such as "Log data events for only two S3 buckets".

            * FieldSelectors (list[dict[str, Any]]): Contains all selector statements in an advanced event selector.

              * Field (str):  A field in an event record on which to filter events to be logged. Supported fields include
                readOnly, eventCategory, eventSource (for management events), eventName, resources.type, and
                resources.ARN.

                * readOnly  - Optional. Can be set to Equals a value of true or false. If you
                  do not add this field, CloudTrail logs both both read and write events. A value of true logs
                  only read events. A value of false logs only write events.

                * eventSource  - For filtering management events only. This can be set only to
                  NotEquals kms.amazonaws.com.

                * eventName  - Can use any operator. You can use it to ﬁlter in or ﬁlter out any data event logged to
                  CloudTrail, such as PutBucket or GetSnapshotBlock. You can have multiple values for this ﬁeld,
                  separated by commas.

                * eventCategory  - This is required. It must be set to Equals, and the value must be Management or Data.

                * resources.type  - This ﬁeld is required. resources.type can only use the Equals operator,
                  and the value can be one of the following:

                  * AWS::S3::Object
                  * AWS::Lambda::Function
                  * AWS::DynamoDB::Table
                  * AWS::S3Outposts::Object
                  * AWS::ManagedBlockchain::Node
                  * AWS::S3ObjectLambda::AccessPoint
                  * AWS::EC2::Snapshot
                  * AWS::S3::AccessPoint
                  * AWS::DynamoDB::Stream
                  * AWS::Glue::Table

                  You can have only one resources.type ﬁeld per selector.
                  To log data events on more than one resource type, add another selector.

                  * resources.ARN  - You can use any operator with resources.ARN, but if you use
                    Equals or NotEquals, the value must exactly match the ARN of a valid resource of the type you've
                    speciﬁed in the template as the value of resources.type. For example, if resources.type equals
                    AWS::S3::Object, the ARN must be in one of the following formats. To log all data events for all
                    objects in a specific S3 bucket, use the StartsWith operator, and include only the bucket ARN as
                    the matching value. The trailing slash is intentional; do not exclude it. Replace the text
                    between less than and greater than symbols (<>) with resource-specific information.

                    * arn:<partition>:s3:::<bucket_name>/
                    * arn:<partition>:s3:::<bucket_name>/<object_path>/

                    When resources.type equals AWS::S3::AccessPoint, and the operator is set to Equals or NotEquals,
                    the ARN must be in one of the following formats. To log events on all objects in an S3 access
                    point, we recommend that you use only the access point ARN, don’t include the object path, and
                    use the StartsWith or NotStartsWith operators.

                    * arn:<partition>:s3:<region>:<account_ID>:accesspoint/<access_point_name>
                    * arn:<partition>:s3:<region>:<account_ID>:accesspoint/<access_point_name>/object/<object_path>

                    When resources.type equals AWS::Lambda::Function, and the operator is set to Equals or
                    NotEquals, the ARN must be in the following format:

                    * arn:<partition>:lambda:<region>:<account_ID>:function:<function_name>

                    When resources.type equals AWS::DynamoDB::Table, and the operator is set to Equals or NotEquals, the ARN must be in
                    the following format:

                    * arn:<partition>:dynamodb:<region>:<account_ID>:table/<table_name>

                    When resources.type equals AWS::S3Outposts::Object, and the operator is set to Equals or
                    NotEquals, the ARN must be in the following format:

                    * arn:<partition>:s3-outposts:<region>:<account_ID>:<object_path>

                    When resources.type equals AWS::ManagedBlockchain::Node, and the operator is set to Equals or
                    NotEquals, the ARN must be in the following format:

                    * arn:<partition>:managedblockchain:<region>:<account_ID>:nodes/<node_ID>

                    When resources.type equals AWS::S3ObjectLambda::AccessPoint, and the operator is set to Equals
                    or NotEquals, the ARN must be in the following format:

                    * arn:<partition>:s3-object-lambda:<region>:<account_ID>:accesspoint/<access_point_name>

                    When resources.type equals AWS::EC2::Snapshot, and the operator is set to Equals or NotEquals,
                    the ARN must be in the following format:

                    * arn:<partition>:ec2:<region>::snapshot/<snapshot_ID>

                    When resources.type equals AWS::DynamoDB::Stream, and the operator is set to Equals or NotEquals,
                    the ARN must be in the following format:

                    * arn:<partition>:dynamodb:<region>:<account_ID>:table/<table_name>/stream/<date_time>

                    When resources.type equals AWS::Glue::Table, and the operator is set to Equals or NotEquals, the ARN
                    must be in the following format:

                    * arn:<partition>:glue:<region>:<account_ID>:table/<database_name>/<table_name>

                * Equals (list[str], Optional):  An operator that includes events that match the exact value of the
                  event record field specified as the value of Field. This is the only valid operator that you can use
                  with the readOnly, eventCategory, and resources.type fields.

                * StartsWith (list[str], Optional):
                  An operator that includes events that match the first few characters of the event record field
                  specified as the value of Field.

                * EndsWith (list[str], Optional):
                  An operator that includes events that match the last few characters of the event record field
                  specified as the value of Field.

                * NotEquals (list[str], Optional):
                  An operator that excludes events that match the exact value of the event record field specified
                  as the value of Field.

                * NotStartsWith (list[str], Optional):
                  An operator that excludes events that match the first few characters of the event record field
                  specified as the value of Field.

                * NotEndsWith (list[str], Optional):
                  An operator that excludes events that match the last few characters of the event record field
                  specified as the value of Field.

    Request Syntax:
        .. code-block:: sls

            [trail_name]:
              aws.cloudtrail.trail.present:
              - resource_id: 'string'
              - s3_bucket_name: 'string'
              - s3_key_prefix: 'string'
              - sns_topic_name: 'string'
              - include_global_service_events: boolean
              - is_multi_region_trail: boolean
              - enable_log_file_validation: boolean
              - cloud_watch_logs_loggroup_arn: 'string'
              - cloud_watch_logs_role_arn: 'string'
              - kms_key_id: 'string'
              - is_organization_trail: boolean
              - tags:
                 - Key: 'string'
                  Value: 'string'

    Examples:
        .. code-block:: sls

            test-trail:
              aws.cloudtrail.trail.present:
              - name: test-trail
              - resource_id: test-trail
              - enable_logfile_validation: false
              - s3_bucket_name: test-bucket1
              - s3_key_prefix: test-bucket
              - sns_topic_name: arn:aws:sns:us-east-2:123456789012:MyTopic
              - include_global_service_events: true
              - is_multi_region_trail: true
              - cloud_watch_logs_loggroup_arn: arn:aws:logs:us-east-2:123456789012:log-group:aws-cloudtrail-logs-123456789012:*
              - cloud_watch_logs_role_arn: arn:aws:iam::123456789012:role/service-role/cloudtrailrole
              - kms_key_id: arn:aws:kms:us-east-2:123456789012:key/12345678-1234-1234-1234-123456789012
              - is_organization_trail: false
              - trail_arn: arn:aws:cloudtrail:us-east-2:123456789012:trail/test-trail
              - tags:
                - Key: trail
                  Value: test-trail
              - is_logging: true
              - event_selectors: null
              - advanced_event_selectors:
                - FieldSelectors:
                  - Field: eventCategory
                    Equals:
                     - Data
                  - Field: resources.type
                    Equals:
                     - AWS::S3::Object
                - Name: Management events selector
                  FieldSelectors:
                  - Field: eventCategory
                    Equals:
                    - Management
              - insight_selectors:
                - InsightType: ApiCallRateInsight
                - InsightType: ApiErrorRateInsight
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    resource_updated = False
    before = None
    if resource_id:
        try:
            before = await hub.exec.boto3.client.cloudtrail.get_trail(
                ctx, Name=resource_id
            )
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = f"{e.__class__.__name__}: {e}"
            result["result"] = False
            return result
    if isinstance(tags, List):
        tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)

    if before:
        try:
            result[
                "old_state"
            ] = await hub.tool.aws.cloudtrail.conversion_utils.convert_raw_cloudtrail_to_present(
                ctx, raw_resource=before["ret"]["Trail"], idem_resource_name=name
            )
            plan_state = copy.deepcopy(result["old_state"])

            update_ret = await hub.tool.aws.cloudtrail.trail.update_trail(
                ctx,
                before=before["ret"]["Trail"],
                s3_bucket_name=s3_bucket_name,
                s3_key_prefix=s3_key_prefix,
                sns_topic_name=sns_topic_name,
                include_global_service_events=include_global_service_events,
                is_multi_region_trail=is_multi_region_trail,
                enable_logfile_validation=enable_logfile_validation,
                cloud_watch_logs_loggroup_arn=cloud_watch_logs_loggroup_arn,
                cloud_watch_logs_role_arn=cloud_watch_logs_role_arn,
                kms_key_id=kms_key_id,
                is_organization_trail=is_organization_trail,
            )
            result["comment"] = result["comment"] + update_ret["comment"]
            result["result"] = update_ret["result"]
            resource_updated = resource_updated or bool(update_ret["ret"])
            if update_ret["ret"] and ctx.get("test", False):
                plan_state = hub.tool.aws.cloudtrail.conversion_utils.update_plan_state(
                    plan_state=plan_state, update_ret=update_ret
                )

            if tags is not None:
                try:
                    old_tags = result["old_state"].get("tags")

                    if tags != old_tags:
                        # Update tags
                        update_ret = await hub.tool.aws.cloudtrail.trail.update_tags(
                            ctx,
                            resource_id=before["ret"]["Trail"]["TrailARN"],
                            old_tags=old_tags,
                            new_tags=tags,
                        )
                        result["result"] = result["result"] and update_ret["result"]
                        result["comment"] = result["comment"] + update_ret["comment"]
                        resource_updated = resource_updated or bool(
                            update_ret["result"]
                        )
                        if ctx.get("test", False) and update_ret["ret"] is not None:
                            plan_state["tags"] = update_ret["ret"]
                except hub.tool.boto3.exception.ClientError as e:
                    result["comment"] = result["comment"] + (
                        f"{e.__class__.__name__}: {e}"
                    )
            if not resource_updated:
                result["comment"] = result["comment"] + (f"{name} already exists",)
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}")
            result["result"] = False
    else:
        if ctx.get("test", False):
            result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                enforced_state={},
                desired_state={
                    "name": name,
                    "s3_bucket_name": s3_bucket_name,
                    "s3_key_prefix": s3_key_prefix,
                    "sns_topic_name": sns_topic_name,
                    "include_global_service_events": include_global_service_events,
                    "is_multi_region_trail": is_multi_region_trail,
                    "enable_logfile_validation": enable_logfile_validation,
                    "cloud_watch_logs_loggroup_arn": cloud_watch_logs_loggroup_arn,
                    "cloud_watch_logs_role_arn": cloud_watch_logs_role_arn,
                    "kms_key_id": kms_key_id,
                    "is_organization_trail": is_organization_trail,
                    "tags": tags,
                    "is_logging": is_logging,
                    "insight_selectors": insight_selectors,
                    "event_selectors": event_selectors,
                    "advanced_event_selectors": advanced_event_selectors,
                },
            )
            result["comment"] = f"Would create aws.cloudtrail.trail {name}"
            return result

        try:
            ret = await hub.exec.boto3.client.cloudtrail.create_trail(
                ctx,
                Name=name,
                S3BucketName=s3_bucket_name,
                S3KeyPrefix=s3_key_prefix,
                SnsTopicName=sns_topic_name,
                IncludeGlobalServiceEvents=include_global_service_events,
                IsMultiRegionTrail=is_multi_region_trail,
                EnableLogFileValidation=enable_logfile_validation,
                CloudWatchLogsLogGroupArn=cloud_watch_logs_loggroup_arn,
                CloudWatchLogsRoleArn=cloud_watch_logs_role_arn,
                KmsKeyId=kms_key_id,
                IsOrganizationTrail=is_organization_trail,
                TagsList=hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags)
                if tags
                else None,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            result["comment"] = result["comment"] + (f"Created '{name}'",)
            resource_id = ret["ret"]["Name"]
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}")
            result["result"] = False
            return result

    ##Update Scenarios for Logging, event selector and insights
    try:
        if before:
            resource = before
        else:
            resource = await hub.exec.boto3.client.cloudtrail.get_trail(
                ctx, Name=resource_id
            )
        update_trail_attr = await hub.tool.aws.cloudtrail.trail.update_trail_attributes(
            ctx,
            before=resource["ret"]["Trail"],
            resource_id=resource["ret"]["Trail"]["TrailARN"],
            is_logging=is_logging,
            update_insight_selectors=insight_selectors,
            update_event_selectors=event_selectors,
            update_advanced_event_selectors=advanced_event_selectors,
        )
        result["comment"] = result["comment"] + update_trail_attr["comment"]
        result["result"] = result["result"] and update_trail_attr["result"]
        resource_updated = resource_updated or bool(update_trail_attr["ret"])
        if update_trail_attr["ret"] and ctx.get("test", False):
            plan_state = hub.tool.aws.cloudtrail.conversion_utils.update_plan_state(
                plan_state=plan_state, update_ret=update_trail_attr
            )

    except hub.tool.boto3.exception.ClientError as e:
        result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}")
        result["result"] = False
        return result

    try:
        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif (not before) or resource_updated:
            after = await hub.exec.boto3.client.cloudtrail.get_trail(
                ctx, Name=resource_id
            )

            result[
                "new_state"
            ] = await hub.tool.aws.cloudtrail.conversion_utils.convert_raw_cloudtrail_to_present(
                ctx, raw_resource=after["ret"]["Trail"], idem_resource_name=name
            )
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
) -> Dict[str, Any]:
    r"""Deletes a trail.

    This operation must be called from the region in which the trail was created. DeleteTrail
    cannot be called on the shadow trails (replicated trails in other regions) of a trail that is enabled in all
    regions.

    Args:
        name(str):
            An Idem name of the resource.

        resource_id(str, Optional):
            Trail name to identify the resource. Idem automatically considers this resource
            being absent if this field is not specified.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            resource_is_absent:
              aws.cloudtrail.trail.absent:
                - name: value
                - resource_id: value
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.cloudtrail.trail", name=name
        )
        return result
    try:
        before = await hub.exec.boto3.client.cloudtrail.get_trail(ctx, Name=resource_id)

        if not before:
            result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
                resource_type="aws.cloudtrail.trail", name=name
            )
        elif ctx.get("test", False):
            result[
                "old_state"
            ] = await hub.tool.aws.cloudtrail.conversion_utils.convert_raw_cloudtrail_to_present(
                ctx, raw_resource=before["ret"]["Trail"], idem_resource_name=name
            )
            result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
                resource_type="aws.cloudtrail.trail", name=name
            )
            return result
        else:
            result[
                "old_state"
            ] = await hub.tool.aws.cloudtrail.conversion_utils.convert_raw_cloudtrail_to_present(
                ctx, raw_resource=before["ret"]["Trail"], idem_resource_name=name
            )
            ret = await hub.exec.boto3.client.cloudtrail.delete_trail(
                ctx, Name=resource_id
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                result["result"] = False
                return result
            result["comment"] = hub.tool.aws.comment_utils.delete_comment(
                resource_type="aws.cloudtrail.trail", name=name
            )

    except hub.tool.boto3.exception.ClientError as e:
        result["comment"] = (f"{e.__class__.__name__}: {e}",)
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""Lists trails that are in the current account including Logging Status, Event Selectors and Insight Selectors as well.

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe aws.cloudtrail.trail
    """
    result = {}

    try:
        ret = await hub.exec.boto3.client.cloudtrail.list_trails(ctx)

        if not ret["result"]:
            hub.log.debug(f"Could not describe trail {ret['comment']}")
            return {}

        trail_list = [trail["TrailARN"] for trail in ret["ret"]["Trails"]]
        key_details = await hub.exec.boto3.client.cloudtrail.describe_trails(
            ctx, trailNameList=trail_list, includeShadowTrails=True
        )

        for trail in key_details["ret"]["trailList"]:
            trail_name = trail["Name"]

            resource_translated = await hub.tool.aws.cloudtrail.conversion_utils.convert_raw_cloudtrail_to_present(
                ctx, raw_resource=trail, idem_resource_name=trail_name
            )
            result[trail_name] = {
                "aws.cloudtrail.trail.present": [
                    {parameter_key: parameter_value}
                    for parameter_key, parameter_value in resource_translated.items()
                ]
            }
    except hub.tool.boto3.exception.ClientError as e:
        result["comment"] = (f"{e.__class__.__name__}: {e}",)

    return result



import json
from typing import Any
from typing import Dict

"""
Util functions for converting raw attributes to present form and vice versa
"""


ATTRIBUTES_NAMES_RAW_TO_PRESENT = {
    "DelaySeconds": "delay_seconds",
    "MaximumMessageSize": "maximum_message_size",
    "MessageRetentionPeriod": "message_retention_period",
    "Policy": "policy",
    "ReceiveMessageWaitTimeSeconds": "receive_message_wait_time_seconds",
    "RedrivePolicy": "redrive_policy",
    "VisibilityTimeout": "visibility_timeout",
    "KmsMasterKeyId": "kms_master_key_id",
    "KmsDataKeyReusePeriodSeconds": "kms_data_key_reuse_period_seconds",
    "SqsManagedSseEnabled": "sqs_managed_sse_enabled",
    "FifoQueue": "fifo_queue",
    "ContentBasedDeduplication": "content_based_deduplication",
    "DeduplicationScope": "deduplication_scope",
    "FifoThroughputLimit": "fifo_throughput_limit",
    "QueueArn": "arn",
}

PRESENT_ATTRIBUTES_TYPES = {
    "delay_seconds": int,
    "maximum_message_size": int,
    "message_retention_period": int,
    "policy": Dict,
    "receive_message_wait_time_seconds": int,
    "redrive_policy": Dict,
    "visibility_timeout": int,
    "kms_master_key_id": str,
    "kms_data_key_reuse_period_seconds": int,
    "sqs_managed_sse_enabled": bool,
    "fifo_queue": bool,
    "content_based_deduplication": bool,
    "deduplication_scope": str,
    "fifo_throughput_limit": str,
    "arn": str,
}


def convert_raw_attributes_to_present(
    hub, raw_attributes: Dict[str, str]
) -> Dict[str, Any]:
    """
    Converts raw attributes to a dict of present values. The types of the values are also changed from string
        to whatever the type of the attribute should be in the present SLS form

    Args:
        hub: required for functions in hub
        raw_attributes(Dict): the attributes as directly returned by the boto3 API

    Returns: a dict of the attributes converted to present form
    """

    present_attributes = {}

    for (
        raw_attribute_name,
        present_attribute_name,
    ) in ATTRIBUTES_NAMES_RAW_TO_PRESENT.items():
        value = raw_attributes.get(raw_attribute_name)

        if value:
            # Convert string values to their correct present value types
            attribute_value_type = PRESENT_ATTRIBUTES_TYPES.get(present_attribute_name)
            if not isinstance(value, attribute_value_type):
                if attribute_value_type == bool:
                    value = value.lower() == "true"
                elif attribute_value_type == int:
                    value = int(value)
                elif attribute_value_type == Dict and value != "":
                    value = json.loads(value)
            present_attributes[present_attribute_name] = value

    return present_attributes


def convert_present_attributes_to_raw(
    hub,
    delay_seconds: int,
    maximum_message_size: int,
    message_retention_period: int,
    policy: Dict,
    receive_message_wait_time_seconds: int,
    redrive_policy: Dict,
    visibility_timeout: int,
    kms_master_key_id: str,
    kms_data_key_reuse_period_seconds: int,
    sqs_managed_sse_enabled: bool,
    fifo_queue: bool,
    content_based_deduplication: bool,
    deduplication_scope: str,
    fifo_throughput_limit: str,
) -> Dict[str, str]:
    """
    Converts present attributes to a dict of raw string values

    Args:
        hub: required for functions in hub
        delay_seconds(int): The length of time, in seconds, for which the delivery of all messages in the queue is delayed
        maximum_message_size(int): The limit of how many bytes a message can contain before Amazon SQS rejects it
        message_retention_period(int): The length of time, in seconds, for which Amazon SQS retains a message
        policy(Dict): The queue's policy
        receive_message_wait_time_seconds(int): The length of time, in seconds, for which a ReceiveMessage
            action waits for a message to arrive
        redrive_policy(Dict): The dict that includes the parameters for the dead-letter queue functionality
            of the source queue
        visibility_timeout(int): The visibility timeout for the queue, in seconds
        kms_master_key_id(str): The ID of an Amazon Web Services managed customer master key (CMK) for
            Amazon SQS or a custom CMK
        kms_data_key_reuse_period_seconds(int): The length of time, in seconds, for which Amazon SQS can
            reuse a data key to encrypt or decrypt messages before calling KMS again
        sqs_managed_sse_enabled(bool, optional): enables server-side queue encryption using SQS owned encryption keys.
            Only one server-side encryption option is supported per queue.
        fifo_queue(bool): Designates a queue as FIFO
        content_based_deduplication(bool): Enables content-based deduplication
        deduplication_scope(str): Specifies whether message deduplication occurs at the message group or queue level
        fifo_throughput_limit(str): Specifies whether the FIFO queue throughput quota applies to the entire

    Returns: a dict of the attributes converted to raw form
    """

    all_attributes = {
        "DelaySeconds": delay_seconds,
        "MaximumMessageSize": maximum_message_size,
        "MessageRetentionPeriod": message_retention_period,
        "Policy": policy,
        "ReceiveMessageWaitTimeSeconds": receive_message_wait_time_seconds,
        "RedrivePolicy": redrive_policy,
        "VisibilityTimeout": visibility_timeout,
        "KmsMasterKeyId": kms_master_key_id,
        "KmsDataKeyReusePeriodSeconds": kms_data_key_reuse_period_seconds,
        "SqsManagedSseEnabled": sqs_managed_sse_enabled,
        "FifoQueue": fifo_queue,
        "ContentBasedDeduplication": content_based_deduplication,
        "DeduplicationScope": deduplication_scope,
        "FifoThroughputLimit": fifo_throughput_limit,
    }

    raw_attributes = {}

    # Remove the None attributes and convert all values to string
    for key, value in all_attributes.items():
        if value is not None:
            # Policy and RedrivePolicy are dict so use json library to convert them to string
            if (key == "Policy" or key == "RedrivePolicy") and not isinstance(
                value, str
            ):
                raw_attributes[key] = json.dumps(value)
            else:
                raw_attributes[key] = str(value)

    return raw_attributes



"""State module for managing AWS CloudWatch Metric Alarms."""
import copy
from dataclasses import field
from dataclasses import make_dataclass
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]

RESOURCE_TYPE = "aws.cloudwatch.metric_alarm"


async def present(
    hub,
    ctx,
    name: str,
    period: int,
    evaluation_periods: int,
    comparison_operator: str,
    resource_id: str = None,
    alarm_description: str = None,
    actions_enabled: bool = True,
    ok_actions: List[str] = None,
    alarm_actions: List[str] = None,
    insufficient_data_actions: List[str] = None,
    metric_name: str = None,
    namespace: str = None,
    statistic: str = None,
    extended_statistic: str = None,
    dimensions: List[
        make_dataclass("Dimension", [("Name", str), ("Value", str)])
    ] = None,
    unit: str = None,
    datapoints_to_alarm: int = None,
    threshold: float = None,
    treat_missing_data: str = None,
    evaluate_low_sample_count_percentile: str = None,
    metrics: List[
        make_dataclass(
            "MetricDataQuery",
            [
                ("Id", str),
                (
                    "MetricStat",
                    make_dataclass(
                        "MetricStat",
                        [
                            (
                                "Metric",
                                make_dataclass(
                                    "Metric",
                                    [
                                        ("Namespace", str, field(default=None)),
                                        ("MetricName", str, field(default=None)),
                                        (
                                            "Dimensions",
                                            List[
                                                make_dataclass(
                                                    "Dimension",
                                                    [("Name", str), ("Value", str)],
                                                )
                                            ],
                                            field(default=None),
                                        ),
                                    ],
                                ),
                            ),
                            ("Period", int),
                            ("Stat", str),
                            ("Unit", str, field(default=None)),
                        ],
                    ),
                    field(default=None),
                ),
                ("Expression", str, field(default=None)),
                ("Label", str, field(default=None)),
                ("ReturnData", bool, field(default=None)),
                ("Period", int, field(default=None)),
                ("AccountId", str, field(default=None)),
            ],
        )
    ] = None,
    threshold_metric_id: str = None,
    tags: Dict[str, Any]
    or List[
        make_dataclass(
            "Tag",
            [("Key", str, field(default=None)), ("Value", str, field(default=None))],
        )
    ] = None,
):
    r"""Creates or updates an alarm and associates it with the specified metric, metric math expression, or anomaly detection model.

    Alarms based on anomaly detection models cannot have Auto Scaling actions.

    When this operation creates an alarm, the alarm state is immediately set to INSUFFICIENT_DATA. The alarm is then
    evaluated and its state is set appropriately. Any actions associated with the new state are then executed.

    When you update an existing alarm, its state is left unchanged, but the update completely overwrites the previous
    configuration of the alarm.

    If you are an IAM user, you must have Amazon EC2 permissions for some alarm operations:
      * The iam:CreateServiceLinkedRole for all alarms with EC2 actions
      * The iam:CreateServiceLinkedRole to create an alarm with Systems Manager OpsItem actions.

    The first time you create an alarm in the Amazon Web Services Management Console, the CLI, or by using the
    PutMetricAlarm API, CloudWatch creates the necessary service-linked role for you. The service-linked roles are
    called AWSServiceRoleForCloudWatchEvents and AWSServiceRoleForCloudWatchAlarms_ActionSSM.

    Args:
        name(str):
            The name for the alarm. This name must be unique within the Region.

        resource_id(str):
            The AWS name of the metric alarm.

        alarm_description(str, Optional):
            The description for the alarm.

        actions_enabled(bool, Optional):
            Indicates whether actions should be executed during any changes to the alarm state. The default is TRUE.

        ok_actions(list, Optional):
            The actions to execute when this alarm transitions to an OK state from any other state.
            Each action is specified as an Amazon Resource Name (ARN).

            Valid Values:
                * arn:aws:automate:*region*:ec2:stop
                * arn:aws:automate:*region*:ec2:terminate
                * arn:aws:automate:*region*:ec2:recover
                * arn:aws:automate:*region*:ec2:reboot
                * arn:aws:sns:*region*:*account-id*:*sns-topic-name*
                * arn:aws:autoscaling:*region*:*account-id*:scalingPolicy:*policy-id*:autoScalingGroupName/*group-friendly-name*:policyName/*policy-friendly-name*
            Valid Values (for use with IAM roles):
                * arn:aws:swf:*region*:*account-id*:action/actions/AWS_EC2.InstanceId.Stop/1.0
                * arn:aws:swf:*region*:*account-id*:action/actions/AWS_EC2.InstanceId.Terminate/1.0
                * arn:aws:swf:*region*:*account-id*:action/actions/AWS_EC2.InstanceId.Reboot/1.0
                * arn:aws:swf:*region*:*account-id*:action/actions/AWS_EC2.InstanceId.Recover/1.0

        alarm_actions(list, Optional):
            The actions to execute when this alarm transitions to the ALARM state from any other state.
            Each action is specified as an Amazon Resource Name (ARN).

            Valid Values:
                * arn:aws:automate:*region*:ec2:stop
                * arn:aws:automate:*region*:ec2:terminate
                * arn:aws:automate:*region*:ec2:recover
                * arn:aws:automate:*region*:ec2:reboot
                * arn:aws:sns:*region*:*account-id*:*sns-topic-name*
                * arn:aws:autoscaling:*region*:*account-id*:scalingPolicy:*policy-id*:autoScalingGroupName/*group-friendly-name*:policyName/*policy-friendly-name*
                * arn:aws:ssm:*region*:*account-id*:opsitem:*severity*
                * arn:aws:ssm-incidents:*region*:*account-id*:response-plan:*response-plan-name*

            Valid Values (for use with IAM roles):
                * arn:aws:swf:*region*:*account-id*:action/actions/AWS_EC2.InstanceId.Stop/1.0
                * arn:aws:swf:*region*:*account-id*:action/actions/AWS_EC2.InstanceId.Terminate/1.0
                * arn:aws:swf:*region*:*account-id*:action/actions/AWS_EC2.InstanceId.Reboot/1.0
                * arn:aws:swf:*region*:*account-id*:action/actions/AWS_EC2.InstanceId.Recover/1.0

        insufficient_data_actions(list, Optional):
            The actions to execute when this alarm transitions to the INSUFFICIENT_DATA state from any other state.
            Each action is specified as an Amazon Resource Name (ARN).

            Valid Values:
                * arn:aws:automate:*region*:ec2:stop
                * arn:aws:automate:*region*:ec2:terminate
                * arn:aws:automate:*region*:ec2:recover
                * arn:aws:automate:*region*:ec2:reboot
                * arn:aws:sns:*region*:*account-id*:*sns-topic-name*
                * arn:aws:autoscaling:*region*:*account-id*:scalingPolicy:*policy-id*:autoScalingGroupName/*group-friendly-name*:policyName/*policy-friendly-name*

            Valid Values (for use with IAM roles):
                * arn:aws:swf:*region*:*account-id*:action/actions/AWS_EC2.InstanceId.Stop/1.0
                * arn:aws:swf:*region*:*account-id*:action/actions/AWS_EC2.InstanceId.Terminate/1.0
                * arn:aws:swf:*region*:*account-id*:action/actions/AWS_EC2.InstanceId.Reboot/1.0

        metric_name(str, Optional):
            The name for the metric associated with the alarm. For each PutMetricAlarm operation, you must specify
            either MetricName or a Metrics array.

        namespace(str, Optional):
            The namespace for the metric associated specified in MetricName.

        statistic(str, Optional):
            The statistic for the metric specified in MetricName , other than percentile. For percentile statistics,
            use ExtendedStatistic. When you call PutMetricAlarm and specify a MetricName, you must specify either
            Statistic or ExtendedStatistic, but not both.

        extended_statistic(str, Optional):
            The percentile statistic for the metric specified in MetricName. Specify a value between p0.0 and p100.
            When you call PutMetricAlarm and specify a MetricName, you must specify either Statistic or
            ExtendedStatistic, but not both.

        dimensions(list[dict[str, Any]], Optional):
            The dimensions for the metric specified in MetricName. Defaults to None.

            * Name (str): The name of the dimension. Dimension names must contain only ASCII characters, must include at
              least one non-whitespace character, and cannot start with a colon (:).

            * Value (str): The value of the dimension. Dimension values must contain only ASCII characters and must
              include at least one non-whitespace character.

        period(int):
            The length, in seconds, used each time the metric specified in MetricName is evaluated.
            Valid values are 10, 30, and any multiple of 60.

        unit(str, Optional):
            The unit of measure for the statistic. For example, the units for the Amazon EC2 NetworkIn metric are
            Bytes because NetworkIn tracks the number of bytes that an instance receives on all network interfaces.
            You can also specify a unit when you create a custom metric. Units help provide conceptual meaning to your
            data. Metric data points that specify a unit of measure, such as Percent, are aggregated separately.

        evaluation_periods(int):
            The number of periods over which data is compared to the specified threshold.
            If you are setting an alarm that requires that a number of consecutive data points be breaching to trigger
            the alarm, this value specifies that number. If you are setting an "M out of N" alarm, this value is the N.

            An alarm's total current evaluation period can be no longer than one day, so this number multiplied by
            Period cannot be more than 86,400 seconds.

        datapoints_to_alarm(int, Optional):
            The number of data points that must be breaching to trigger the alarm.
            This is used only if you are setting an "M out of N" alarm. In that case, this value is the M.

        threshold(float, Optional):
            The value against which the specified statistic is compared.

            This parameter is required for alarms based on static thresholds, but should not be used for alarms based
            on anomaly detection models.

        comparison_operator(str):
            The arithmetic operation to use when comparing the specified statistic and threshold. The specified
            statistic value is used as the first operand.

            The values LessThanLowerOrGreaterThanUpperThreshold, LessThanLowerThreshold, and GreaterThanUpperThreshold
            are used only for alarms based on anomaly detection models.

        treat_missing_data(str, Optional):
            Sets how this alarm is to handle missing data points. If TreatMissingData is omitted, the default behavior
            of missing is used.

            Valid Values: breaching | notBreaching | ignore | missing

        evaluate_low_sample_count_percentile(str, Optional):
            Used only for alarms based on percentiles. If you specify ignore, the alarm state does not change during
            periods with too few data points to be statistically significant. If you specify evaluate or omit this
            parameter, the alarm is always evaluated and possibly changes state no matter how many data points are
            available.

            Valid Values: evaluate | ignore

        metrics(list[dict[str, Any]], Optional):
            An array of MetricDataQuery structures that enable you to create an alarm based on the result of a metric
            math expression. For each PutMetricAlarm operation, you must specify either MetricName or a Metrics array.

            Each item in the Metrics array either retrieves a metric or performs a math expression.

            One item in the Metrics array is the expression that the alarm watches. You
            designate this expression by setting ReturnData to true for this object in the array. For more
            information, see MetricDataQuery.

            If you use the Metrics parameter, you cannot include the MetricName, Dimensions, Period, Namespace,
            Statistic, or ExtendedStatistic parameters of PutMetricAlarm in the same operation. Instead, you retrieve
            the metrics you are using in your math expression as part of the Metrics array.

            Defaults to None.

            * Id (str):
                A short name used to tie this object to the results in the response. This name must be unique
                within a single call to GetMetricData. If you are performing math expressions on this set of
                data, this name represents that data and can serve as a variable in the mathematical expression.
                The valid characters are letters, numbers, and underscore. The first character must be a
                lowercase letter.

            * MetricStat (dict[str, Any], Optional):
                The metric to be returned, along with statistics, period, and units. Use this parameter only if
                this object is retrieving a metric and not performing a math expression on returned data.

                Within one MetricDataQuery object, you must specify either Expression or MetricStat but not both.

                * Metric (dict[str, Any]): The metric to return, including the metric name, namespace, and dimensions.
                    * Namespace (str, Optional): The namespace of the metric.

                    * MetricName (str, Optional): The name of the metric. This is a required field.

                    * Dimensions (list[dict[str, Any]], Optional): The dimensions for the metric.
                        * Name (str): The name of the dimension. Dimension names must contain only ASCII characters,
                          must include at least one non-whitespace character, and cannot start with a colon (:).

                        * Value (str): The value of the dimension. Dimension values must contain only ASCII characters
                          and must include at least one non-whitespace character.

                * Period (int):
                    The granularity, in seconds, of the returned data points. For metrics with regular resolution, a
                    period can be as short as one minute (60 seconds) and must be a multiple of 60. For high-resolution
                    metrics that are collected at intervals of less than one minute, the period can be 1, 5, 10, 30,
                    60, or any multiple of 60. High-resolution metrics are those metrics stored by a PutMetricData call
                    that includes a StorageResolution of 1 second.

                    If the StartTime parameter specifies a time stamp that is greater than 3 hours ago, you must
                    specify the period as follows or no data points in that time range is returned:

                    * Start time between 3 hours and 15 days ago - Use a multiple of 60 seconds (1 minute).
                    * Start time between 15 and 63 days ago - Use a multiple of 300 seconds (5 minutes).
                    * Start time greater than 63 days ago - Use a multiple of 3600 seconds (1 hour).

                * Stat (str):
                    The statistic to return. It can include any CloudWatch statistic or extended statistic.

                * Unit (str, Optional):
                    When you are using a Put operation, this defines what unit you want to use when storing the
                    metric.

                    In a Get operation, if you omit Unit then all data that was collected with any unit is returned,
                    along with the corresponding units that were specified when the data was reported to CloudWatch. If
                    you specify a unit, the operation returns only data that was collected with that unit specified. If
                    you specify a unit that does not match the data collected, the results of the operation are null.
                    CloudWatch does not perform unit conversions.

            * Expression (str, Optional):
                This field can contain either a Metrics Insights query, or a metric math expression to be
                performed on the returned data. For more information about Metrics Insights queries, see Metrics
                Insights query components and syntax in the Amazon CloudWatch User Guide.

                A math expression can use the ID of the other metrics or queries to refer to those metrics, and can
                also use the ID of other expressions to use the result of those expressions. For more information about
                metric math expressions, see Metric Math Syntax and Functions in the Amazon CloudWatch User Guide.

                Within each MetricDataQuery object, you must specify either Expression or MetricStat but not both.

            * Label (str, Optional):
                A human-readable label for this metric or expression. This is especially useful if this is an
                expression, so that you know what the value represents. If the metric or expression is shown in
                a CloudWatch dashboard widget, the label is shown. If Label is omitted, CloudWatch generates a
                default.

                You can put dynamic expressions into a label, so that it is more descriptive. For more information,
                see Using Dynamic Labels.

            * ReturnData (bool, Optional):
                When used in GetMetricData, this option indicates whether to return the timestamps and raw data
                values of this metric. If you are performing this call just to do math expressions and do not
                also need the raw data returned, you can specify False. If you omit this, the default of True is used.

                When used in PutMetricAlarm, specify True for the one expression result to use as the alarm. For all
                other metrics and expressions in the same PutMetricAlarm operation, specify ReturnData as False.

            * Period (int, Optional):
                The granularity, in seconds, of the returned data points. For metrics with regular resolution, a
                period can be as short as one minute (60 seconds) and must be a multiple of 60. For high-resolution
                metrics that are collected at intervals of less than one minute, the period can be 1, 5, 10, 30, 60,
                or any multiple of 60. High-resolution metrics are those metrics stored by a PutMetricData operation
                that includes a StorageResolution of 1 second.

            * AccountId (str, Optional):
                The ID of the account where the metrics are located, if this is a cross-account alarm.

                Use this field only for PutMetricAlarm operations. It is not used in GetMetricData operations.

        tags(list or dict, Optional):
            A List of tags in the format of [{"Key": tag-key, "Value": tag-value}] or dict in the format of
            {tag-key: tag-value} to associate with the alarm. You can associate as many as 50 tags with
            an alarm.

            Tags can help you organize and categorize your resources. You can also use them to scope user permissions
            by granting a user permission to access or change only resources with certain tag values.

            If you are using this operation to update an existing alarm, any tags you specify in this parameter are
            ignored. To change the tags of an existing alarm, use TagResource or UntagResource. Defaults to None.

            * (Key): A string that you can use to assign a value. The combination of tag keys and values can help you
              organize and categorize your resources.

            * (Value): The value for the specified tag key.

        threshold_metric_id(str):
            If this is an alarm based on an anomaly detection model, make this value match the ID of the
            ANOMALY_DETECTION_BAND function.

            If your alarm uses this parameter, it cannot have Auto Scaling actions.

    Request Syntax:
        .. code-block:: sls

            [metric-alarm-name]:
              aws.cloudwatch.metric_alarm.present:
                - name: 'string'
                - alarm_description: 'string'
                - actions_enabled: True|False
                - ok_actions: ['string']
                - alarm_actions: ['string']
                - insufficient_data_actions: ['string']
                - metric_name: 'string'
                - namespace: 'string'
                - statistic: 'string'
                - dimensions: 'list'
                - period: 'integer'
                - unit: 'string'
                - evaluation_periods: 'integer'
                - datapoints_to_alarm: 'integer'
                - threshold: 'integer'
                - comparison_operator: 'string'
                - treat_missing_data: 'string'
                - evaluate_low_sample_count_percentile: 'string'
                - metrics: 'list'
                - threshold_metric_id: 'string'
                - tags:
                    - Key: 'string'
                      Value: 'string'

    Returns:
         Dict[str, Any]

    Examples:
        .. code-block:: sls

            awsec2-i-0f36e2b10a7463129-LessThanOrEqualToThreshold-CPUUtilization:
              aws.cloudwatch.metric_alarm.present:
                - name: awsec2-i-0f36e2b10a7463129-LessThanOrEqualToThreshold-CPUUtilization
                - alarm_description: "stop EC2 instance if it utilizes CPU less than 5"
                - actions_enabled: True
                - alarm_actions:
                  - arn:aws:swf:*region*:*account-id*:action/actions/AWS_EC2.InstanceId.Stop/1.0
                - insufficient_data_actions: []
                - metric_name: CPUUtilization
                - namespace: AWS/EC2
                - statistic: Average
                - dimensions:
                  - Name: InstanceId
                    Value: i-0f36e2b10a7463129
                - period: 60
                - evaluation_periods: 1
                - datapoints_to_alarm: 1
                - threshold: 20
                - comparison_operator: LessThanThreshold
                - tags:
                  - Key: type
                    Value: metric_alarm

    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    resource_updated = False
    before = None

    if isinstance(tags, List):
        tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict(tags)

    resource_parameters = {
        "AlarmName": name,
        "AlarmDescription": alarm_description,
        "ActionsEnabled": actions_enabled,
        "OKActions": ok_actions,
        "AlarmActions": alarm_actions,
        "InsufficientDataActions": insufficient_data_actions,
        "MetricName": metric_name,
        "Namespace": namespace,
        "Statistic": statistic,
        "ExtendedStatistic": extended_statistic,
        "Dimensions": dimensions,
        "Period": period,
        "Unit": unit,
        "EvaluationPeriods": evaluation_periods,
        "DatapointsToAlarm": datapoints_to_alarm,
        "Threshold": threshold,
        "ComparisonOperator": comparison_operator,
        "TreatMissingData": treat_missing_data,
        "EvaluateLowSampleCountPercentile": evaluate_low_sample_count_percentile,
        "Metrics": metrics,
        "ThresholdMetricId": threshold_metric_id,
        "Tags": hub.tool.aws.tag_utils.convert_tag_dict_to_list(tags),
    }

    if resource_id:
        resource = hub.tool.boto3.resource.create(
            ctx, "cloudwatch", "Alarm", resource_id
        )
        before = await hub.tool.boto3.resource.describe(resource)

    try:

        if before:

            convert_ret = await hub.tool.aws.cloudwatch.conversion_utils.convert_raw_metric_alarm_to_present(
                ctx, raw_resource=before, idem_resource_name=name
            )
            result["result"] = convert_ret["result"]
            if not result["result"]:
                result["comment"] = result["comment"] + convert_ret["comment"]
            result["old_state"] = convert_ret["ret"]

            plan_state = copy.deepcopy(result["old_state"])
            # Update metric alarm
            update_ret = await hub.tool.aws.cloudwatch.metric_alarm.update_metric(
                ctx,
                alarm_name=name,
                raw_resource=before,
                resource_parameters=resource_parameters,
            )
            result["comment"] = result["comment"] + update_ret["comment"]
            result["result"] = update_ret["result"]
            resource_updated = resource_updated or bool(update_ret["ret"])
            if update_ret["ret"] and ctx.get("test", False):
                for key in [
                    "alarm_description",
                    "actions_enabled",
                    "ok_actions",
                    "alarm_actions",
                    "metric_name",
                    "namespace",
                    "statistic",
                    "dimensions",
                    "period",
                    "evaluation_periods",
                    "datapoints_to_alarm",
                    "threshold",
                    "comparison_operator",
                ]:
                    if key in update_ret["ret"]:
                        plan_state[key] = update_ret["ret"][key]
            if tags is not None and tags != result["old_state"].get("tags"):
                # Update tags
                update_tag_ret = (
                    await hub.tool.aws.cloudwatch.metric_alarm.update_metric_tags(
                        ctx=ctx,
                        alarm_arn=before.get("AlarmArn"),
                        old_tags=result["old_state"].get("tags", {}),
                        new_tags=tags,
                    )
                )
                result["result"] = result["result"] and update_tag_ret["result"]
                result["comment"] = result["comment"] + update_tag_ret["comment"]
                resource_updated = resource_updated or bool(update_tag_ret["result"])

                if ctx.get("test", False) and update_tag_ret["ret"] is not None:
                    plan_state["tags"] = update_tag_ret["ret"]

            if not resource_updated:
                result["comment"] = result["comment"] + (
                    f"aws.cloudwatch.metric_alarm '{name}' has no property need to be updated.",
                )
        else:
            if ctx.get("test", False):
                result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                    enforced_state={},
                    desired_state={
                        "name": name,
                        "alarm_description": alarm_description,
                        "actions_enabled": actions_enabled,
                        "ok_actions": ok_actions,
                        "alarm_actions": alarm_actions,
                        "insufficient_data_actions": insufficient_data_actions,
                        "metric_name": metric_name,
                        "namespace": namespace,
                        "statistic": statistic,
                        "extended_statistic": extended_statistic,
                        "dimensions": dimensions,
                        "period": period,
                        "unit": unit,
                        "evaluation_periods": evaluation_periods,
                        "datapoints_to_alarm": datapoints_to_alarm,
                        "threshold": threshold,
                        "comparison_operator": comparison_operator,
                        "treat_missing_data": treat_missing_data,
                        "evaluate_low_sample_count_percentile": evaluate_low_sample_count_percentile,
                        "metrics": metrics,
                        "tags": tags,
                        "threshold_metric_id": threshold_metric_id,
                    },
                )
                result["comment"] = hub.tool.aws.comment_utils.would_create_comment(
                    resource_type=RESOURCE_TYPE, name=name
                )
                return result

            # Create metric alarm
            ret = await hub.exec.boto3.client.cloudwatch.put_metric_alarm(
                ctx, **resource_parameters
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = result["comment"] + ret["comment"]
                return result
            result["comment"] = result[
                "comment"
            ] + hub.tool.aws.comment_utils.create_comment(RESOURCE_TYPE, name)

        if ctx.get("test", False):
            result["new_state"] = plan_state
        elif (not before) or resource_updated:
            resource = hub.tool.boto3.resource.create(ctx, "cloudwatch", "Alarm", name)
            after = await hub.tool.boto3.resource.describe(resource)
            convert_ret = await hub.tool.aws.cloudwatch.conversion_utils.convert_raw_metric_alarm_to_present(
                ctx, raw_resource=after, idem_resource_name=name
            )
            result["result"] = convert_ret["result"]
            if not result["result"]:
                result["comment"] = result["comment"] + convert_ret["comment"]
            result["new_state"] = convert_ret["ret"]
        else:
            result["new_state"] = copy.deepcopy(result["old_state"])
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False

    return result


async def absent(hub, ctx, name: str, resource_id: str = None) -> Dict[str, Any]:
    r"""Deletes the specified alarms.

    You can delete up to 100 alarms in one operation. However, this total can include no
    more than one composite alarm. For example, you could delete 99 metric alarms and one composite alarms with one
    operation, but you can't delete two composite alarms with one operation.

    Args:
        name(str): An Idem name of the metric alarm.
        resource_id(str, Optional): The AWS name of the metric alarm. Idem automatically considers this resource
         being absent if this field is not specified.

    Request Syntax:
        .. code-block:: sls

            [alarm name]:
               aws.cloudwatch.metric_alarm.absent:
                 - resource_id: value

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            [alarm name]:
               aws.cloudwatch.metric_alarm.absent:
                 - resource_id: value

    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type=RESOURCE_TYPE, name=name
        )
        return result
    resource = hub.tool.boto3.resource.create(ctx, "cloudwatch", "Alarm", resource_id)
    before = await hub.tool.boto3.resource.describe(resource)

    if not before:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type=RESOURCE_TYPE, name=name
        )
    elif ctx.get("test", False):
        convert_ret = await hub.tool.aws.cloudwatch.conversion_utils.convert_raw_metric_alarm_to_present(
            ctx, raw_resource=before, idem_resource_name=name
        )
        result["result"] = convert_ret["result"]
        if not result["result"]:
            result["comment"] = result["comment"] + convert_ret["comment"]
        result["old_state"] = convert_ret["ret"]
        result["comment"] = hub.tool.aws.comment_utils.would_delete_comment(
            resource_type=RESOURCE_TYPE, name=name
        )
        return result
    else:
        try:
            convert_ret = await hub.tool.aws.cloudwatch.conversion_utils.convert_raw_metric_alarm_to_present(
                ctx, raw_resource=before, idem_resource_name=name
            )
            result["result"] = convert_ret["result"]
            if not result["result"]:
                result["comment"] = result["comment"] + convert_ret["comment"]
            result["old_state"] = convert_ret["ret"]
            ret = await hub.exec.boto3.client.cloudwatch.delete_alarms(
                ctx, AlarmNames=[resource_id]
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                result["result"] = False
                return result
            result["comment"] = hub.tool.aws.comment_utils.delete_comment(
                resource_type=RESOURCE_TYPE, name=name
            )

        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = result["comment"] + (f"{e.__class__.__name__}: {e}",)

    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""Retrieves the specified alarms.

    You can filter the results by specifying a prefix for the alarm name, the alarm state, or a prefix for any action.
    To use this operation and return information about composite alarms, you must be signed on with the
    cloudwatch:DescribeAlarms permission that is scoped to * . You can't return information about composite alarms if
    your cloudwatch:DescribeAlarms permission has a narrower scope.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe aws.cloudwatch.metric_alarm
    """
    result = {}
    ret = await hub.exec.boto3.client.cloudwatch.describe_alarms(ctx)

    if not ret["result"]:
        hub.log.debug(f"Could not describe alarm metrics {ret['comment']}")
        return {}

    for resource in ret["ret"]["MetricAlarms"]:
        alarm_name = resource["AlarmName"]
        convert_ret = await hub.tool.aws.cloudwatch.conversion_utils.convert_raw_metric_alarm_to_present(
            ctx, raw_resource=resource, idem_resource_name=alarm_name
        )
        if not convert_ret["result"]:
            hub.log.warning(
                f"Could not describe alarm metrics '{alarm_name}' with error {convert_ret['comment']}"
            )
            continue
        translated_resource = convert_ret["ret"]
        result[translated_resource["resource_id"]] = {
            "aws.cloudwatch.metric_alarm.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in translated_resource.items()
            ]
        }

    return result



from typing import Dict
from typing import List


async def apply(
    hub,
    ctx,
    resource,
    *,
    old_value: Dict[str, str],
    new_value: Dict[str, str],
    comments: List[str],
) -> bool:
    """
    Modify an ec2 instance based on a single parameter in it's "present" state

    Args:
        hub:
        ctx: The ctx from a state module call
        resource: An ec2 instance resource object
        old_value: The previous value from the attributes of an existing instance
        new_value: The desired value from the ec2 instance present state parameters
        comments: A running list of comments abound the update process
    """
    result = True
    tags_to_create = set(new_value.keys()) - set(old_value.keys())
    tags_to_delete = set(old_value.keys()) - set(new_value.keys())

    tags_to_modify = set(old_value.keys()).intersection(set(new_value.keys()))
    for key in tags_to_modify:
        if old_value[key] != new_value[key]:
            # Add the tag to the delete operation then to the create operation with tth new value
            tags_to_delete.add(key)
            tags_to_create.add(key)

    # Delete old tags
    ret = await hub.exec.boto3.client.ec2.delete_tags(
        ctx,
        Resources=[resource.id],
        Tags=[{"Key": key, "Value": old_value[key]} for key in tags_to_delete],
    )
    result &= ret.result
    if ret.comment:
        comments.append(ret.comment)

    # Create new tags
    ret = await hub.exec.boto3.client.ec2.create_tags(
        ctx,
        Resources=[resource.id],
        Tags=[{"Key": key, "Value": new_value[key]} for key in tags_to_create],
    )
    result &= ret.result
    if ret.comment:
        comments.append(ret.comment)

    return result



from collections import OrderedDict
from typing import Any
from typing import Dict

"""
Util functions to convert raw resource state from AWS EKS to present input format.
"""


def convert_raw_addon_to_present(
    hub, raw_resource: Dict[str, Any], idem_resource_name: str = None
) -> Dict[str, Any]:
    resource_id = raw_resource.get("addonName")
    resource_parameters = OrderedDict(
        {
            "clusterName": "cluster_name",
            "addonArn": "addon_arn",
            "addonVersion": "addon_version",
            "releaseVersion": "release_version",
            "status": "status",
            "clientRequestToken": "client_request_token",
            "serviceAccountRoleArn": "service_account_role_arn",
            "tags": "tags",
        }
    )
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            resource_translated[parameter_present] = (
                raw_resource.get(parameter_raw).copy()
                if isinstance(raw_resource.get(parameter_raw), dict)
                else raw_resource.get(parameter_raw)
            )

    return resource_translated


def convert_raw_cluster_to_present(
    hub, raw_resource: Dict[str, Any], idem_resource_name: str = None
) -> Dict[str, Any]:
    resource_id = raw_resource.get("name")
    resource_parameters = OrderedDict(
        {
            "roleArn": "role_arn",
            "arn": "arn",
            "status": "status",
            "version": "version",
            "resourcesVpcConfig": "resources_vpc_config",
            "kubernetesNetworkConfig": "kubernetes_network_config",
            "logging": "logging",
            "encryptionConfig": "encryption_config",
            "clientRequestToken": "client_request_token",
            "tags": "tags",
        }
    )

    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            resource_translated[parameter_present] = (
                raw_resource.get(parameter_raw).copy()
                if isinstance(raw_resource.get(parameter_raw), dict)
                else raw_resource.get(parameter_raw)
            )

    return resource_translated


def convert_raw_nodegroup_to_present(
    hub, raw_resource: Dict[str, Any], idem_resource_name: str = None
) -> Dict[str, Any]:
    resource_id = raw_resource.get("nodegroupName")
    resource_parameters = OrderedDict(
        {
            "clusterName": "cluster_name",
            "nodegroupArn": "nodegroup_arn",
            "version": "version",
            "releaseVersion": "release_version",
            "status": "status",
            "capacityType": "capacity_type",
            "scalingConfig": "scaling_config",
            "instanceTypes": "instance_types",
            "subnets": "subnets",
            "remoteAccess": "remote_access",
            "amiType": "ami_type",
            "nodeRole": "node_role",
            "labels": "labels",
            "taints": "taints",
            "diskSize": "disk_size",
            "updateConfig": "update_config",
            "launchTemplate": "launch_template",
            "tags": "tags",
        }
    )
    resource_translated = {"name": idem_resource_name, "resource_id": resource_id}
    for parameter_raw, parameter_present in resource_parameters.items():
        if parameter_raw in raw_resource:
            resource_translated[parameter_present] = (
                raw_resource.get(parameter_raw).copy()
                if isinstance(raw_resource.get(parameter_raw), dict)
                else raw_resource.get(parameter_raw)
            )

    return resource_translated


def convert_raw_fargate_profile_to_present(
    hub, raw_resource: Dict[str, Any]
) -> Dict[str, Any]:
    describe_parameters = OrderedDict(
        {
            "clusterName": "cluster_name",
            "podExecutionRoleArn": "pod_execution_role_arn",
            "status": "status",
        }
    )
    name = raw_resource.get("fargateProfileName")
    translated_resource = {"name": name, "resource_id": name}
    for parameter_old, parameter_new in describe_parameters.items():
        if parameter_old in raw_resource:
            translated_resource[parameter_new] = raw_resource.get(parameter_old)
        if "subnets" in raw_resource:
            translated_resource["subnets"] = raw_resource.get("subnets").copy()
        if "selectors" in raw_resource:
            translated_resource["selectors"] = raw_resource.get("selectors").copy()
        if "tags" in raw_resource:
            translated_resource["tags"] = raw_resource.get("tags").copy()
    return translated_resource



"""State module for managing Amazon KMS Keys."""
import copy
from typing import Any
from typing import Dict
from typing import List

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    description: str = "",
    key_usage: str = "ENCRYPT_DECRYPT",
    key_spec: str = "SYMMETRIC_DEFAULT",
    key_state: str = "Enabled",
    origin: str = "AWS_KMS",
    multi_region: bool = False,
    policy: str = None,
    bypass_policy_lockout_safety_check: bool = False,
    enable_key_rotation: bool = False,
    tags: List[Dict[str, Any]] or Dict[str, Any] = None,
    timeout: Dict = None,
) -> Dict[str, Any]:
    """Creates an AWS KMS key.

    Update limitations:
      * ``policy`` can be updated, but cannot be cleared once set.
      * ``multi_region``, ``key_usage``, and ``key_spec`` cannot be updated.
      * ``enable_key_rotation`` cannot be enabled on asymmetric KMS keys.

    Args:
        name(str):
            An Idem name of the resource.
        resource_id(str, Optional):
            The ID or ARN of the key in Amazon Web Services.
        description(str, Optional):
            A description of the KMS key.
        key_usage(str, Optional):
            Determines the cryptographic operations for which you can use the KMS key. The default value is ``ENCRYPT_DECRYPT``.
            Valid values are ``ENCRYPT_DECRYPT``, ``GENERATE_VERIFY_MAC``, ``SIGN_VERIFY``.
        key_spec(str, Optional):
            Specifies the type of KMS key to create. The default value is ``SYMMETRIC_DEFAULT``.
            Valid values are ``SYMMETRIC_DEFAULT``, ``HMAC_224``, ``HMAC_256``, ``HMAC_384``, ``HMAC_512``, ``RSA_2048``, ``RSA_3072``, ``RSA_4096``,
            ``ECC_NIST_P256``, ``ECC_NIST_P384``, ``ECC_NIST_P521``, ``ECC_SECG_P256K1``, ``SM2``.
        key_state(str, Optional):
            Whether the key is enabled or not. The default value is ``Enabled``.
            Valid values are ``Enabled``, ``Disabled``.
        origin(str, Optional):
            The source of the key material for the KMS key. The default value is ``AWS_KMS``.
            Valid values are ``AWS_KMS``, ``EXTERNAL``, ``AWS_CLOUDHSM``.
        multi_region(bool, Optional):
            Creates a multi-Region primary key that you can replicate into other Amazon Web Services Regions. Default value is ``False``.
        policy(str, Optional):
            The key policy to attach to the KMS key. If you do not specify a key policy, KMS attaches a default key policy to the KMS key.
        bypass_policy_lockout_safety_check(bool, Optional):
            A flag to indicate whether to bypass the key policy lockout safety check. Default value is ``False``.

            .. warning::
              Setting this value to ``True`` increases the risk that the KMS key becomes unmanageable.
              Do not set this value to ``True`` indiscriminately.

        enable_key_rotation(bool, Optional):
            Whether to enable or disable automatic rotation of the key material of the specified symmetric encryption KMS key.
            Default value is ``False``.
        tags(dict or list, Optional):
            Dict in the format of ``{tag-key: tag-value}`` or List of tags in the format of
            ``[{"TagKey": tag-key, "TagValue": tag-value}]`` to associate with the key.
            To use this parameter, you must have ``kms:TagResource`` permission in an IAM policy.

            * TagKey (*str*):
                The key of the tag.
            * TagValue (*str*):
                The value of the tag.
        timeout(dict, Optional):
            Timeout configuration for update of AWS KMS key.

            * update (*dict, Optional*):
                Timeout configuration when updating a KMS key.

                * delay (*int, Optional*):
                    The amount of time in seconds to wait between attempts. Default value is ``2``.
                * max_attempts (*int, Optional*):
                    Max attempts of waiting for change. Default value is ``120``.

    Request Syntax:
      .. code-block:: sls

        [idem_test_aws_kms_key]:
          aws.kms.key.present:
            - name: 'string'
            - resource_id: 'string'
            - description: 'string'
            - key_usage: 'ENCRYPT_DECRYPT|GENERATE_VERIFY_MAC|SIGN_VERIFY'
            - key_spec: 'SYMMETRIC_DEFAULT|HMAC_224|HMAC_256|HMAC_384|HMAC_512|RSA_2048|RSA_3072|RSA_4096|ECC_NIST_P256|ECC_NIST_P384|ECC_NIST_P521|ECC_SECG_P256K1|SM2'
            - key_state: 'Enabled|Disabled'
            - origin: 'AWS_KMS|EXTERNAL|AWS_CLOUDHSM'
            - multi_region: True|False
            - policy: 'string'
            - bypass_policy_lockout_safety_check: True|False
            - enable_key_rotation: True|False
            - tags:
                - TagKey: 'string'
                  TagValue: 'string'
            - timeout:
                update:
                  delay: int
                  max_attemps: int

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            idem_test_aws_kms_key:
              aws.kms.key.present:
                - name: idem_test_kms_key
                - description: 'AWS KMS KEY'
                - key_usage: 'ENCRYPT_DECRYPT'
                - key_spec: 'SYMMETRIC_DEFAULT'
                - key_state: 'Enabled'
                - origin: 'AWS_KMS'
                - multi_region: False
                - policy:
                    Version: '2012-10-17'
                    Statement:
                      - Sid: 'EnableIAMUserPermissions'
                        Effect: 'Allow'
                        Principal:
                          AWS: 'arn:aws:iam::111122223333:root'
                        Action: ['kms:*']
                        Resource: '*'
                - bypass_policy_lockout_safety_check: False
                - enable_key_rotation: False
                - tags:
                    - TagKey: provider
                      TagValue: idem
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)

    plan_state = {}
    updated_vals = {}
    before = None
    if resource_id is not None:
        before = await hub.exec.boto3.client.kms.describe_key(ctx, KeyId=resource_id)

    if isinstance(tags, List):
        tags = hub.tool.aws.tag_utils.convert_tag_list_to_dict_tagkey(tags)

    policy = hub.tool.aws.state_comparison_utils.standardise_json(policy)

    if before and before["result"]:
        result["comment"] = (f"aws.kms.key '{name}' already exists.",)
        try:
            result[
                "old_state"
            ] = await hub.tool.aws.kms.conversion_utils.convert_raw_key_to_present(
                ctx, raw_resource=before["ret"]["KeyMetadata"]
            )
        except Exception as e:
            result["comment"] = result["comment"] + (str(e),)
            result["result"] = False
            return result

        # Used for 'test'
        plan_state = copy.deepcopy(result["old_state"])

        # Update key tags if tags are specified
        if tags is not None and tags != result["old_state"].get("tags"):
            hub.log.debug(f"aws.kms.key '{name}' tags update")
            # For tag operations key_id is used and not resource_id
            update_ret = await hub.tool.aws.kms.tag.update_tags(
                ctx=ctx,
                key_id=resource_id,
                old_tags=result["old_state"].get("tags", []),
                new_tags=tags,
            )

            result["result"] = update_ret["result"]
            result["comment"] = result["comment"] + update_ret["comment"]
            if not result["result"]:
                return result

            plan_state["tags"] = update_ret["ret"]
            updated_vals["tags"] = tags

        # Enable/Disable key
        # No updates for key_state "PendingDeletion", which means the key is scheduled to be deleted.
        old_state = result["old_state"].get("key_state", "")
        if key_state != old_state:
            if ctx.get("test", False):
                result["comment"] = result["comment"] + (
                    f"Would update state on aws.kms.key '{name}'.",
                )
                plan_state["key_state"] = key_state
            else:
                update_ret = None
                if old_state == "Enabled" and key_state == "Disabled":
                    update_ret = await hub.exec.boto3.client.kms.disable_key(
                        ctx, KeyId=result["old_state"]["resource_id"]
                    )
                elif old_state == "Disabled" and key_state == "Enabled":
                    update_ret = await hub.exec.boto3.client.kms.enable_key(
                        ctx, KeyId=result["old_state"]["resource_id"]
                    )

                if update_ret:
                    hub.log.debug(
                        f"Updated the state of aws.kms.key '{name}' to '{key_state}'."
                    )
                    result["comment"] = result["comment"] + (
                        f"Updated aws.kms.key '{name}' state to '{key_state}'.",
                    )
                    updated_vals["key_state"] = key_state
                else:
                    hub.log.warning(
                        f"Failed to update the state of aws.kms.key '{name}' to '{key_state}'. {update_ret['comment']} "
                    )
                    result["result"] = update_ret["result"]
                    result["comment"] = result["comment"] + update_ret["comment"]
                    return result

        # Update policy and/or bypass flag if needed and if policy is set
        update_policy = (
            policy
            and not hub.tool.aws.state_comparison_utils.is_json_identical(
                result["old_state"].get("policy"), policy
            )
        )
        update_bypass_lockout = bypass_policy_lockout_safety_check != result[
            "old_state"
        ].get("bypass_policy_lockout_safety_check", False)
        if policy and (update_policy or update_bypass_lockout):
            if ctx.get("test", False):
                if update_policy:
                    result["comment"] = result["comment"] + (
                        f"Would update policy of aws.kms.key '{name}'.",
                    )
                if update_bypass_lockout:
                    result["comment"] = result["comment"] + (
                        f"Would update bypass_policy_lockout_safety_check of aws.kms.key '{name}'.",
                    )
                plan_state["policy"] = policy
                plan_state[
                    "bypass_policy_lockout_safety_check"
                ] = bypass_policy_lockout_safety_check
            else:
                update_ret = await hub.exec.boto3.client.kms.put_key_policy(
                    ctx,
                    KeyId=result["old_state"]["resource_id"],
                    Policy=policy,
                    PolicyName="default",
                    BypassPolicyLockoutSafetyCheck=bypass_policy_lockout_safety_check,
                )

                if update_ret["result"]:
                    if update_policy:
                        result["comment"] = result["comment"] + (
                            f"Updated aws.kms.key '{name}' policy.",
                        )
                        updated_vals["policy"] = policy
                    if update_bypass_lockout:
                        result["comment"] = result["comment"] + (
                            f"Updated aws.kms.key '{name}' bypass_policy_lockout_safety_check.",
                        )
                        updated_vals[
                            "bypass_policy_lockout_safety_check"
                        ] = bypass_policy_lockout_safety_check
                else:
                    hub.log.warning(
                        f"Failed to update the policy and/or bypass_policy_lockout_safety_check of aws.kms.key '{name}': {update_ret['comment']}"
                    )
                    result["result"] = update_ret["result"]
                    result["comment"] = result["comment"] + update_ret["comment"]
                    return result

        # Update description if needed
        if description and description != result["old_state"].get("description"):
            if ctx.get("test", False):
                result["comment"] = result["comment"] + (
                    f"Would update description of aws.kms.key '{name}'.",
                )
                plan_state["description"] = description
            else:
                update_ret = await hub.exec.boto3.client.kms.update_key_description(
                    ctx,
                    KeyId=result["old_state"]["resource_id"],
                    Description=description,
                )

                if update_ret["result"]:
                    result["comment"] = result["comment"] + (
                        f"Updated aws.kms.key '{name}' description.",
                    )
                    updated_vals["description"] = description
                else:
                    hub.log.warning(
                        f"Failed to update the description of aws.kms.key '{name}': {update_ret['comment']}"
                    )
                    result["result"] = update_ret["result"]
                    result["comment"] = result["comment"] + update_ret["comment"]
                    return result
    else:
        try:
            if ctx.get("test", False):
                result["new_state"] = hub.tool.aws.test_state_utils.generate_test_state(
                    enforced_state={},
                    desired_state={
                        "name": name,
                        "resource_id": f"key-{name}-resource_id",
                        "arn": f"key-{name}-arn",
                        "description": description,
                        "key_usage": key_usage,
                        "key_spec": key_spec,
                        "origin": origin,
                        "multi_region": multi_region,
                        "policy": policy,
                        "bypass_policy_lockout_safety_check": bypass_policy_lockout_safety_check,
                        "enable_key_rotation": enable_key_rotation,
                        "tags": tags,
                    },
                )
                result["comment"] = (f"Would create aws.kms.key {name}",)
                return result

            ret = await hub.exec.boto3.client.kms.create_key(
                ctx,
                Description=description,
                KeyUsage=key_usage,
                KeySpec=key_spec,
                Origin=origin,
                MultiRegion=multi_region,
                Policy=policy,
                BypassPolicyLockoutSafetyCheck=bypass_policy_lockout_safety_check,
                Tags=hub.tool.aws.tag_utils.convert_tag_dict_to_list_tagkey(tags)
                if tags
                else None,
            )

            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                return result
            resource_id = ret["ret"]["KeyMetadata"]["KeyId"]
            result["comment"] = (
                f"Created aws.kms.key '{name}' with description '{description}'.",
            )
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = f"{e.__class__.__name__}: {e}"
            result["result"] = False

    # Set key rotation:
    # Enable it on a newly created key or Enable/Disable on an existing key,
    # only if the key_state is Enabled
    if (enable_key_rotation and result.get("old_state") is None) or (
        result.get("old_state")
        and (
            key_state == "Enabled"
            or (key_state == None and result["old_state"].get("key_state") == "Enabled")
        )
        and enable_key_rotation != result["old_state"].get("enable_key_rotation", False)
    ):
        update_ret = await hub.tool.aws.kms.key.set_key_rotation(
            ctx=ctx,
            resource_id=resource_id,
            enable_key_rotation=enable_key_rotation,
        )
        plan_state["enable_kms_rotation"] = enable_key_rotation
        result["comment"] = result["comment"] + update_ret["comment"]
        if not update_ret["result"]:
            result["result"] = False
            return result

    if ctx.get("test", False):
        result["new_state"] = plan_state
    else:
        if updated_vals:
            wait_updates_res = await hub.tool.aws.kms.key.wait_for_updates(
                ctx,
                timeout=timeout.get("update") if timeout else None,
                resource_id=resource_id,
                updates=updated_vals,
            )
            if wait_updates_res["result"] is False:
                # Do not fail just add to comments
                result["comment"] = result["comment"] + wait_updates_res["comment"]

            result["new_state"] = wait_updates_res["ret"]
        else:
            try:
                after = await hub.exec.boto3.client.kms.describe_key(
                    ctx, KeyId=resource_id
                )
                if after.get("ret"):
                    result[
                        "new_state"
                    ] = await hub.tool.aws.kms.conversion_utils.convert_raw_key_to_present(
                        ctx, raw_resource=after["ret"].get("KeyMetadata", None)
                    )
            except Exception as e:
                result["comment"] = result["comment"] + (str(e),)
                result["result"] = False

    return result


async def absent(
    hub, ctx, name: str, resource_id: str = None, pending_window_in_days: int = 7
) -> Dict[str, Any]:
    """Deletes an AWS KMS key.

    Key cannot be immediately deleted but can be scheduled to be deleted.
    Once the key is set to be deleted in ``pending_window_in_days`` a deletion date is set on the key and it cannot be modified.
    So deleting again with a different ``pending_window_in_days`` is ignored. Also key can be disabled using the "present" function
    with ``key_state: 'Disabled'``.

    Args:
        name(str):
            An Idem name of the resource.
        resource_id(str, Optional):
            The ID or ARN of the key in Amazon Web Services.

            .. warning::
              Idem automatically considers this resource being absent if this field is not specified.

        pending_window_in_days(int, Optional):
            The waiting period, specified in number of days. After the waiting period ends, KMS deletes the KMS key. Default value is ``7``.

    Returns:
        Dict[str, Any]

    Request Syntax:
      .. code-block:: sls

        [idem_test_aws_kms_key]:
          aws.kms.key.absent:
            - name: 'string'
            - resource_id: 'string'
            - pending_window_in_days: 'int'

    Examples:
        .. code-block:: sls

            idem_test_aws_kms_key:
              aws.kms.key.absent:
                - name: idem_test_kms_key
                - resource_id: 1234abcd-12ab-34cd-56ef-1234567890ab
                - pending_window_in_days: 2
    """
    result = dict(comment=(), old_state=None, new_state=None, name=name, result=True)
    if not resource_id:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.kms.key", name=name
        )
        return result
    before = await hub.exec.boto3.client.kms.describe_key(ctx, KeyId=resource_id)

    if not before:
        result["comment"] = hub.tool.aws.comment_utils.already_absent_comment(
            resource_type="aws.kms.key", name=name
        )
        return result
    else:
        try:
            result[
                "old_state"
            ] = await hub.tool.aws.kms.conversion_utils.convert_raw_key_to_present(
                ctx, raw_resource=before["ret"]["KeyMetadata"]
            )
        except Exception as e:
            result["comment"] = (str(e),)
            result["result"] = False
            return result

        if before["ret"]["KeyMetadata"].get("DeletionDate", None):
            result["comment"] = (
                f"aws.kms.key '{resource_id}' already scheduled to be deleted in '{pending_window_in_days}' days",
            )
            return result

        if ctx.get("test", False):
            result["comment"] = (
                f"Would schedule deletion of aws.kms.key '{resource_id}' in {pending_window_in_days} days",
            )
            return result

        try:
            # Minimum deletion schedule 7 days (from 7 - 30)
            ret = await hub.exec.boto3.client.kms.schedule_key_deletion(
                ctx,
                KeyId=resource_id,
                PendingWindowInDays=pending_window_in_days,
            )
            result["result"] = ret["result"]
            if not result["result"]:
                result["comment"] = ret["comment"]
                result["result"] = False
                return result
            result["comment"] = (
                f"aws.kms.key '{resource_id}' is scheduled for deletion in {pending_window_in_days} days",
            )
        except hub.tool.boto3.exception.ClientError as e:
            result["comment"] = (f"{e.__class__.__name__}: {e}",)

    try:
        after = await hub.exec.boto3.client.kms.describe_key(
            ctx, KeyId=(resource_id if resource_id else name)
        )
        if after.get("ret"):
            result[
                "new_state"
            ] = await hub.tool.aws.kms.conversion_utils.convert_raw_key_to_present(
                ctx, raw_resource=after["ret"]["KeyMetadata"]
            )
    except Exception as e:
        result["comment"] = result["comment"] + (str(e),)
        result["result"] = False
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    """Describes AWS KMS keys in a way that can be recreated/managed with the corresponding "present" function.

    Returns:
        Dict[str, Dict[str, Any]]

    Examples:
        .. code-block:: bash

            $ idem describe aws.kms.key
    """
    result = {}
    ret = await hub.exec.boto3.client.kms.list_keys(ctx)

    if not ret["result"]:
        hub.log.debug(f"Could not describe aws.kms.keys {ret['comment']}")
        return {}

    for key in ret["ret"]["Keys"]:
        key_details = await hub.exec.boto3.client.kms.describe_key(
            ctx, KeyId=key.get("KeyId", None)
        )
        if key_details.get("result") is False:
            hub.log.debug(
                f"Failed to describe key with Id {key.get('KeyId')}: {key_details.get('comment')}"
            )
            continue

        try:
            translated_resource = (
                await hub.tool.aws.kms.conversion_utils.convert_raw_key_to_present(
                    ctx, raw_resource=key_details["ret"]["KeyMetadata"]
                )
            )
        except Exception as e:
            hub.log.debug(f"Failed to convert key with Id {key.get('KeyId')}: {e}")
            continue

        result[translated_resource["resource_id"]] = {
            "aws.kms.key.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in translated_resource.items()
            ]
        }

    return result
